# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from re import compile as Re

import shutil
import time

from dns import flags, name, rdataclass, rdatatype
from dns.edns import EDECode

import pytest

from isctest.util import param

import isctest
import isctest.mark

pytestmark = pytest.mark.extra_artifacts(
    [
        "*/K*",
        "*/dsset-*",
        "*/*.bk",
        "*/*.conf",
        "*/*.db",
        "*/*.id",
        "*/*.jnl",
        "*/*.jbk",
        "*/*.key",
        "*/*.signed",
        "*/settime.out.*",
        "ans*/ans.run",
        "*/trusted.keys",
        "*/*.bad",
        "*/*.next",
        "*/*.stripped",
        "*/*.tmp",
        "*/*.stage?",
        "*/*.patched",
        "*/*.lower",
        "*/*.upper",
        "*/*.unsplit",
    ]
)


def getfrom(file):
    with open(file, encoding="utf-8") as f:
        return f.read().strip()


@pytest.mark.requires_zones_loaded("ns2", "ns3")
@pytest.mark.parametrize(
    "qname, qtype",
    [
        param("a.example.", "A"),
        param("rfc2535.example.", "SOA"),
    ],
)
def test_load_transfer(qname, qtype):
    # check that we can load and transfer zone
    msg = isctest.query.create(qname, qtype)
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.3")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res1)


def test_insecure_rrsig():
    # check that for a rrsig query against a validating resolver where the
    # authoritative zone is unsigned (insecure delegation), noerror is
    # returned.
    msg = isctest.query.create("a.insecure.example", "RRSIG")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.rr_count_eq(res.answer, 0)
    isctest.check.rr_count_eq(res.authority, 1)
    isctest.check.rr_count_eq(res.additional, 0)
    assert str(res.authority[0].name) == "insecure.example."
    assert res.authority[0].rdtype == rdatatype.SOA


def test_adflag():
    # compare auth and recursive answers
    msg = isctest.query.create("a.example", "A", dnssec=False)
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)

    # check no AD flag in authoritative response
    isctest.check.noadflag(res1)

    # check validating resolver sends AD=1 if the client sent AD=1
    isctest.check.adflag(res2)

    # check that AD=0 unless the client sent AD=1
    msg = isctest.query.create("a.example", "A", dnssec=False, ad=False)
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noadflag(res2)


def test_secure_root(ns4, default_algorithm):
    # check that a query for a secure root validates
    msg = isctest.query.create(".", "KEY")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)

    # check that "rndc secroots" dumps the trusted keys
    key = int(getfrom("ns1/managed.key.id"))
    response = ns4.rndc("secroots -")
    assert f"./{default_algorithm.name}/{key} ; static" in response.out
    assert len(response.out.splitlines()) == 10


def test_positive_validation_nsec():
    # positive answer
    msg = isctest.query.create("a.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)

    # wildcard
    msg = isctest.query.create("a.wild.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.adflag(res2)

    assert str(res2.authority[0].name) == "*.wild.example."
    assert res2.authority[0].rdtype == rdatatype.NSEC
    nsecs = [str(a).split(" ", maxsplit=1)[0] for a in res2.authority[0]]
    assert "z.example." in nsecs
    assert res2.authority[1].rdtype == rdatatype.RRSIG
    assert res2.authority[1].covers == rdatatype.NSEC

    # mixed case
    for rrtype in ["a", "txt", "aaaa", "loc"]:
        msg = isctest.query.create("mixedcase.secure.example", rrtype)
        res1 = isctest.query.tcp(msg, "10.53.0.3")
        res2 = isctest.query.tcp(msg, "10.53.0.4")
        isctest.check.same_answer(res1, res2)
        isctest.check.noerror(res2)
        isctest.check.adflag(res2)


def test_positive_validation_nsec3():
    # positive answer
    msg = isctest.query.create("a.nsec3.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.adflag(res2)

    # wildcard
    msg = isctest.query.create("a.wild.nsec3.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.adflag(res2)
    isctest.check.rr_count_eq(res2.authority, 2)

    # unknown NSEC3 hash algorithm
    msg = isctest.query.create("nsec3-unknown.example", "SOA", dnssec=False)
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res1)
    isctest.check.noerror(res2)
    isctest.check.adflag(res2)
    isctest.check.rr_count_eq(res2.answer, 1)


def test_positive_validation_dname_at_apex():
    # an apex DNAME is signed by the DNSKEY living at the DNAME owner
    # name itself; fetching that key must not be mistaken for a
    # non-advancing alias chain (GL #6176)
    msg = isctest.query.create("a.dname-at-apex-nsec3.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)
    answers = {(str(rr.name), rr.rdtype) for rr in res.answer}
    assert ("dname-at-apex-nsec3.example.", rdatatype.DNAME) in answers
    assert ("a.example.", rdatatype.A) in answers


def test_positive_validation_optout():
    # positive answer
    msg = isctest.query.create("a.optout.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)

    # wildcard
    msg = isctest.query.create("a.wild.optout.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.noadflag(res2)

    # unknown NSEC3 hash algorithm
    msg = isctest.query.create("optout-unknown.example", "SOA", dnssec=False)
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res1)
    isctest.check.noerror(res2)
    isctest.check.adflag(res2)
    isctest.check.rr_count_eq(res2.answer, 1)


def answer_has(r, rdtype):
    return bool([r for r in r.answer if r.rdtype == rdtype])


def test_chain_validation():
    # check validation of ANY response
    msg = isctest.query.create("foo.example", "ANY")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.rr_count_eq(res2.answer, 6)  # 2 records, 1 NSEC, 3 RRSIGs

    # check validation of CNAME response
    msg = isctest.query.create("cname1.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.rr_count_eq(res2.answer, 4)  # CNAME, TXT, 2 RRSIGs

    # check validation of DNAME response
    msg = isctest.query.create("foo.dname1.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.rr_count_eq(res2.answer, 5)  # DNAME, TXT, 2 RRSIGs, synth CNAME

    # check validation of CNAME response to ANY query
    msg = isctest.query.create("cname2.example", "ANY")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.rr_count_eq(res2.answer, 4)  # CNAME, NSEC, 2 RRSIGs

    # check validation of DNAME response to ANY query
    msg = isctest.query.create("foo.dname2.example", "ANY")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.rr_count_eq(res2.answer, 3)  # DNAME, RSRIG, synth CNAME

    # check bad CNAME signature is caught after +CD query
    msg = isctest.query.create("bad-cname.example", "A", dnssec=False, cd=True)
    # query once with CD to prime the cache
    res = isctest.query.tcp(msg, "10.53.0.4")
    # query again with CD, bogus pending data should be returned
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    assert "a.example." in str(res.answer[0])
    assert "10.0.0.1" in str(res.answer[1])
    # query again without CD, bogus data should be rejected
    msg = isctest.query.create("bad-cname.example", "A", dnssec=False)
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.servfail(res)

    # check bad DNAME signature is caught after +CD query
    msg = isctest.query.create("a.bad-dname.example", "A", dnssec=False, cd=True)
    # query once with CD to prime the cache
    res = isctest.query.tcp(msg, "10.53.0.4")
    # query again with CD, bogus pending data should be returned
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    assert "example." in str(res.answer[0])
    assert "a.example." in str(res.answer[1])
    assert "10.0.0.1" in str(res.answer[2])
    # query again without CD, bogus data should be rejected
    msg = isctest.query.create("a.bad-dname.example", "A", dnssec=False)
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.servfail(res)

    # check DNSKEY lookup via CNAME
    msg = isctest.query.create("cnameandkey.secure.example", "DNSKEY")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)
    assert answer_has(res2, rdatatype.CNAME)

    # check DNSKEY lookup via DNAME
    msg = isctest.query.create("a.dnameandkey.secure.example", "DNSKEY")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)
    assert answer_has(res2, rdatatype.DNAME)

    # check KEY lookup via DNAME
    msg = isctest.query.create("a.dnameandkey.secure.example", "KEY")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)
    assert answer_has(res2, rdatatype.DNAME)


@isctest.mark.rsasha1
def test_signing_algorithms_rsasha1():
    # rsasha1 (should work with FIPS mode we're as only validating)
    msg = isctest.query.create("a.rsasha1.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)

    # rsasha1 (1024 bits) NSEC
    msg = isctest.query.create("a.rsasha1-1024.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)


def test_signing_algorithms():
    # rsasha256
    msg = isctest.query.create("a.rsasha256.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)

    # rsasha512
    msg = isctest.query.create("a.rsasha512.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)

    # KSK-only DNSKEY
    msg = isctest.query.create("a.kskonly.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)


def test_private_algorithms(ns4):
    # positive answer, private algorithm
    msg = isctest.query.create("a.rsasha256oid.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)

    # positive answer, unknown private algorithm
    msg = isctest.query.create("a.unknownoid.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noadflag(res2)

    # positive answer, extra ds for private algorithm
    msg = isctest.query.create("a.extradsoid.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)

    # positive anwer, extra ds for unknown private algorithm
    with ns4.watch_log_from_here() as watcher:
        msg = isctest.query.create("a.extradsunknownoid.example", "A")
        res1 = isctest.query.tcp(msg, "10.53.0.3")
        res2 = isctest.query.tcp(msg, "10.53.0.4")
        isctest.check.noerror(res1)
        isctest.check.servfail(res2)
        watcher.wait_for_line("no DNSKEY matching DS")


@isctest.mark.extended_ds_digest
def test_private_algorithms_extended_ds():
    # check positive validation with extra ds using extended digest
    # type for unknown private algorithm
    msg = isctest.query.create("a.extended-ds-unknown-oid.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)


def test_negative_validation_nsec():
    # nxdomain
    msg = isctest.query.create("q.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.nxdomain(res2)
    isctest.check.adflag(res2)

    # nodata
    msg = isctest.query.create("a.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.empty_answer(res2)
    isctest.check.adflag(res2)

    # negative wildcard
    msg = isctest.query.create("b.wild.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.adflag(res2)


def test_negative_validation_nsec3():
    # nxdomain
    msg = isctest.query.create("q.nsec3.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.nxdomain(res2)
    isctest.check.adflag(res2)

    # nodata
    msg = isctest.query.create("a.nsec3.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.empty_answer(res2)
    isctest.check.adflag(res2)

    # negative wildcard
    msg = isctest.query.create("b.wild.nsec3.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.adflag(res2)

    # check NSEC3 zone with mismatched NSEC3PARAM / NSEC parameters
    msg = isctest.query.create("non-exist.badparam", "A")
    res = isctest.query.tcp(msg, "10.53.0.2")
    isctest.check.nxdomain(res)

    # check negative unknown NSEC3 hash algorithm does not validate
    msg = isctest.query.create("nsec3-unknown.example", "A", dnssec=False)
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res1)
    isctest.check.servfail(res2)


def test_excessive_nsec3_iterations(ns2, ns3):
    msg = "zone too-many-iterations/IN: excessive NSEC3PARAM iterations"
    assert msg in ns2.log
    assert msg in ns3.log

    # check fallback to insecure with NSEC3 iterations is too high
    msg = isctest.query.create("does-not-exist.too-many-iterations", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_data(res1, res2)
    isctest.check.noadflag(res2)
    isctest.check.rr_count_eq(res2.answer, 0)
    isctest.check.rr_count_eq(res2.authority, 8)

    # check fallback to insecure with NSEC3 iterations is too high (nodata)
    msg = isctest.query.create("a.too-many-iterations", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_data(res1, res2)
    isctest.check.noadflag(res2)
    isctest.check.rr_count_eq(res2.answer, 0)
    isctest.check.rr_count_eq(res2.authority, 4)

    # check fallback to insecure with NSEC3 iterations is too high (wildcard)
    msg = isctest.query.create("wild.a.too-many-iterations", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.section_equal(res1.answer, res2.answer)
    isctest.check.noadflag(res2)
    isctest.check.rr_count_eq(res2.answer, 2)
    isctest.check.rr_count_eq(res2.authority, 2)
    a, _ = res2.answer
    assert str(a.name) == "wild.a.too-many-iterations."
    assert str(a[0]) == "10.0.0.3"

    # check fallback to insecure with high NSEC3 iterations (wildcard nodata)
    msg = isctest.query.create("wild.a.too-many-iterations", 100)
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_data(res1, res2)
    isctest.check.noadflag(res2)
    isctest.check.rr_count_eq(res2.authority, 8)


def test_auth_nsec3():
    # nxdomain response, closest encloser with 0 empty non-terminals
    msg = isctest.query.create("b.b.b.b.b.a.nsec3.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.nxdomain(res)
    # closest encloser (a.nsec3.example):
    rrset = res.get_rrset(
        res.authority,
        name.from_text("6OVDUHTN094ML2PV8AN90U0DPU823GH2.nsec3.example."),
        rdataclass.IN,
        rdatatype.NSEC3,
    )
    assert rrset, "NSEC3 missing from AUTHORITY: " + str(res)
    assert "7AT0S0RIDCJRFF2M5H5AAV22CSFJBUL4" in str(rrset[0]).upper()
    # no QNAME (b.a.nsec3.example/DSPF4R9UKOEPJ9O34E1H4539LSOTL14E)
    rrset = res.get_rrset(
        res.authority,
        name.from_text("BEJ5GMQA872JF4DAGQ0R3O5Q7A2O5S9L.nsec3.example."),
        rdataclass.IN,
        rdatatype.NSEC3,
    )
    assert rrset, "expected NSEC3 missing from AUTHORITY: " + str(res)
    assert "EF2S05SGK1IR2K5SKMFIRERGQCLMR18M" in str(rrset[0]).upper()
    # no WILDCARD (*.a.nsec3.example/TFGQ60S97BS31IT1EBEDO63ETM0T5JFA)
    rrset = res.get_rrset(
        res.authority,
        name.from_text("R8EVDMNIGNOKME4LH2H90OSP2PRSNJ1Q.nsec3.example."),
        rdataclass.IN,
        rdatatype.NSEC3,
    )
    assert rrset, "expected NSEC3 missing from AUTHORITY: " + str(res)
    assert "VH656EQUD4J02OFVSO4GKOK5D02MS1TL" in str(rrset[0]).upper()

    # nxdomain response, closest encloser with 1 ENT
    msg = isctest.query.create("b.b.b.b.b.a.a.nsec3.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.nxdomain(res)
    # closest encloser (a.a.nsec3.example):
    rrset = res.get_rrset(
        res.authority,
        name.from_text("NGCJFSOLJUUE27PFNQNJIME4TQ0OU2DH.nsec3.example."),
        rdataclass.IN,
        rdatatype.NSEC3,
    )
    assert rrset, "expected NSEC3 missing from AUTHORITY: " + str(res)
    assert "R8EVDMNIGNOKME4LH2H90OSP2PRSNJ1Q" in str(rrset[0]).upper()
    # noqname (b.a.a.nsec3.example):
    rrset = res.get_rrset(
        res.authority,
        name.from_text("R8EVDMNIGNOKME4LH2H90OSP2PRSNJ1Q.nsec3.example."),
        rdataclass.IN,
        rdatatype.NSEC3,
    )
    assert rrset, "expected NSEC3 missing from AUTHORITY: " + str(res)
    assert "VH656EQUD4J02OFVSO4GKOK5D02MS1TL" in str(rrset[0]).upper()
    # no wildcard (*.a.a.nsec3.example/V7JNNDJ4NLRIU195FRB7DLUCSLU4LLFM)
    # is covered by the noqname proof in this case

    # nxdomain response, closest encloser with 2 ENTs
    msg = isctest.query.create("b.b.b.b.b.a.a.a.nsec3.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.nxdomain(res)
    # closest encloser (a.a.a.nsec3.example):
    rrset = res.get_rrset(
        res.authority,
        name.from_text("H7RHPDCHSVVRAND332F878C8AB6IBJQV.nsec3.example."),
        rdataclass.IN,
        rdatatype.NSEC3,
    )
    assert rrset, "expected NSEC3 missing from AUTHORITY: " + str(res)
    assert "K8IG76R2UPQ13IKFO49L7IB9JRVB6QJI" in str(rrset[0]).upper()
    # noqname (b.a.a.a.nsec3.example/18Q8D89RM8GGRSSOPFRB05QS6VEGB1P4)
    rrset = res.get_rrset(
        res.authority,
        name.from_text("0T7VH688AEK0612T69V8692OCMJD50M4.nsec3.example."),
        rdataclass.IN,
        rdatatype.NSEC3,
    )
    assert rrset, "expected NSEC3 missing from AUTHORITY: " + str(res)
    assert "1HARMGSKJH0EBU2EI2OJIKTDPIQA6KBI" in str(rrset[0]).upper()
    # no WILDCARD (*.a.a.a.nsec3.example/8113LDMSEFPUAG4VGFF1C8KLOUT4Q6PH)
    rrset = res.get_rrset(
        res.authority,
        name.from_text("7AT0S0RIDCJRFF2M5H5AAV22CSFJBUL4.nsec3.example."),
        rdataclass.IN,
        rdatatype.NSEC3,
    )
    assert rrset, "expected NSEC3 missing from AUTHORITY: " + str(res)
    assert "BEJ5GMQA872JF4DAGQ0R3O5Q7A2O5S9L" in str(rrset[0]).upper()


def test_negative_validation_optout():
    # nxdomain
    msg = isctest.query.create("q.optout.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.nxdomain(res2)
    isctest.check.noadflag(res2)

    # nodata
    msg = isctest.query.create("a.optout.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.empty_answer(res2)
    isctest.check.adflag(res2)

    # negative wildcard
    msg = isctest.query.create("b.wild.optout.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.noadflag(res2)

    # empty NODATA
    msg = isctest.query.create("empty.optout.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.noadflag(res2)

    # (rt22007 regression tests:)
    # check optout NSEC3 referral with only insecure delegatons
    msg = isctest.query.create("delegation.single-nsec3", "A")
    res = isctest.query.tcp(msg, "10.53.0.2")
    isctest.check.noerror(res)
    isctest.check.has_authority(res)
    for rrset in res.authority:
        if (
            rrset.rdtype != rdatatype.NSEC3
            or "3KL3NK1HKQ4IUEEHBEF12VGFKUETNBAN" not in rrset.name
        ):
            continue
        assert "1 1 1 - 3KL3NK1HKQ4IUEEHBEF12VGFKUETNBAN" in str(rrset[0])

    # check optout NSEC3 NXDOMAIN with only insecure delegatons
    msg = isctest.query.create("nonexist.single-nsec3", "A")
    res = isctest.query.tcp(msg, "10.53.0.2")
    isctest.check.nxdomain(res)
    isctest.check.has_authority(res)
    for rrset in res.authority:
        if (
            rrset.rdtype != rdatatype.NSEC3
            or "3KL3NK1HKQ4IUEEHBEF12VGFKUETNBAN" not in rrset.name
        ):
            continue
        assert "1 1 1 - 3KL3NK1HKQ4IUEEHBEF12VGFKUETNBAN" in str(rrset[0])

    # check optout NSEC3 NODATA with only insecure delegatons
    msg = isctest.query.create("single-nsec3", "A")
    res = isctest.query.tcp(msg, "10.53.0.2")
    isctest.check.noerror(res)
    isctest.check.has_authority(res)
    for rrset in res.authority:
        if (
            rrset.rdtype != rdatatype.NSEC3
            or "3KL3NK1HKQ4IUEEHBEF12VGFKUETNBAN" not in rrset.name
        ):
            continue
        assert "1 1 1 - 3KL3NK1HKQ4IUEEHBEF12VGFKUETNBAN" in str(rrset[0])

    # check negative unknown NSEC3-OPTOUT hash algorithm does not validate
    msg = isctest.query.create("optout-unknown.example", "A", dnssec=False)
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res1)
    isctest.check.servfail(res2)


def test_cache(ns4):
    ns4.rndc("dumpdb -cache")
    with isctest.log.WatchLogFromStart("ns4/named_dump.db") as watcher:
        watcher.wait_for_all(
            [
                "; key id = ",  # check that key id's are logged when dumping the cache
                "; example. RRSIG NSEC ",  # check for RRSIG covered type in negative cache
            ]
        )

    # check validated data are not cached longer than originalttl
    msg = isctest.query.create("a.ttlpatch.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.rr_count_eq(res1.answer, 2)
    isctest.check.rr_count_eq(res2.answer, 2)
    for rrset in res1.answer:
        assert 3000 <= rrset.ttl <= 3600
    for rrset in res2.answer:
        assert rrset.ttl <= 300

    # first query for a NS record, to cache NSEC and RRSIG(NSEC)
    msg = isctest.query.create("normalthenrrsig.secure.example", "NS")
    isctest.query.tcp(msg, "10.53.0.4")
    # query for a record, then follow it with a query for the
    # corresponding RRSIG, check that it's answered from the cache
    msg = isctest.query.create("normalthenrrsig.secure.example", "A")
    isctest.query.tcp(msg, "10.53.0.4")

    msg = isctest.query.create("normalthenrrsig.secure.example", "RRSIG")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.raflag(res2)

    # check direct query for RRSIG: if it's not cached with other records,
    # it should result in an empty response.
    msg = isctest.query.create("rrsigonly.secure.example", "RRSIG")
    res1 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.empty_answer(res1)
    isctest.check.noraflag(res1)

    # check that a DNSKEY query with no data still gets cached
    msg = isctest.query.create("insecure.example", "DNSKEY")
    res1 = isctest.query.tcp(msg, "10.53.0.4")
    time.sleep(1)  # give the TTL time to change
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    if res1.authority[0].ttl == res2.authority[0].ttl:
        time.sleep(1)
        res2 = isctest.query.tcp(msg, "10.53.0.4")
        assert res1.authority[0].ttl != res2.authority[0].ttl


def test_insecure_proof_nsec(ns4):
    # 1-server positive
    msg = isctest.query.create("a.insecure.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.noadflag(res2)

    # 1-server negative
    msg = isctest.query.create("q.insecure.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.nxdomain(res2)
    isctest.check.noadflag(res2)

    # 1-server negative with SOA hack
    msg = isctest.query.create("r.insecure.example", "SOA")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.nxdomain(res2)
    isctest.check.noadflag(res2)
    assert res2.authority[0].rdtype == rdatatype.SOA
    assert res2.authority[0].ttl == 0

    # 2-server positive
    msg = isctest.query.create("a.insecure.secure.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.noadflag(res2)

    # 2-server negative
    msg = isctest.query.create("q.insecure.secure.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.nxdomain(res2)
    isctest.check.noadflag(res2)

    # 2-server negative with SOA hack
    msg = isctest.query.create("r.insecure.secure.example", "SOA")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.nxdomain(res2)
    isctest.check.noadflag(res2)

    # insecurity proof using negative cache
    ns4.rndc("flush")
    msg = isctest.query.create("insecure.example", "DS", cd=True)
    isctest.query.tcp(msg, "10.53.0.4")

    def query_and_check_nxdomain():
        msg = isctest.query.create("nonexistent.insecure.example", "A")
        res = isctest.query.tcp(msg, "10.53.0.4")
        isctest.check.nxdomain(res)
        return True

    isctest.run.retry_with_timeout(query_and_check_nxdomain, 20)

    # check insecure negative response with an unsigned NSEC
    # first try the auth server...
    msg = isctest.query.create("nsec-rrsigs-stripped", "TXT")
    res1 = isctest.query.udp(msg, "10.53.0.10")
    isctest.check.noerror(res1)
    isctest.check.empty_answer(res1)
    isctest.check.rr_count_eq(res1.authority, 2)
    isctest.check.rr_count_eq(res1.additional, 0)
    # make sure there's no RRSIG(NSEC)
    for rrset in res1.authority:
        assert rrset.rdtype != rdatatype.RRSIG or rrset.covers != rdatatype.NSEC
    # now try the resolver
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_data(res1, res2)
    isctest.check.noadflag(res2)


def test_insecure_proof_nsec3():
    # 1-server
    msg = isctest.query.create("a.insecure.nsec3.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_data(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.noadflag(res2)

    # 1-server negative
    msg = isctest.query.create("q.insecure.nsec3.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_data(res1, res2)
    isctest.check.nxdomain(res2)
    isctest.check.noadflag(res2)

    # 1-server negative with SOA hack
    msg = isctest.query.create("r.insecure.nsec3.example", "SOA")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_data(res1, res2)
    isctest.check.nxdomain(res2)
    isctest.check.noadflag(res2)
    assert res2.authority[0].rdtype == rdatatype.SOA
    assert res2.authority[0].ttl == 0


def test_insecure_proof_optout():
    # 1-server
    msg = isctest.query.create("a.insecure.optout.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_data(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.noadflag(res2)

    # 1-server negative
    msg = isctest.query.create("q.insecure.optout.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_data(res1, res2)
    isctest.check.nxdomain(res2)
    isctest.check.noadflag(res2)

    # 1-server negative with SOA hack
    msg = isctest.query.create("r.insecure.optout.example", "SOA")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_data(res1, res2)
    isctest.check.nxdomain(res2)
    isctest.check.noadflag(res2)
    assert res2.authority[0].rdtype == rdatatype.SOA
    assert res2.authority[0].ttl == 0


def test_below_cname():
    # check insecure zone below a cname resolves
    msg = isctest.query.create("insecure.below-cname.example", "SOA")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)
    isctest.check.rr_count_eq(res.answer, 1)

    # check secure zone below a cname resolves and validates
    msg = isctest.query.create("secure.below-cname.example", "SOA")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)
    isctest.check.rr_count_eq(res.answer, 2)


@pytest.mark.parametrize(
    "qname",
    [
        "a.secure.example",  # NSEC/NSEC
        "a.nsec3.example",  # NSEC/NSEC3
        "a.optout.example",  # NSEC/OPTOUT
        "a.secure.nsec3.example",  # NSEC3/NSEC
        "a.nsec3.nsec3.example",  # NSEC3/NSEC3
        "a.optout.nsec3.example",  # NSEC3/OPTOUT
        "a.secure.optout.example",  # OPTOUT/NSEC
        "a.nsec3.optout.example",  # OPTOUT/NSEC3
        "a.optout.optout.example",  # OPTOUT/OPTOUT
    ],
)
def test_positive_validation_multistage(qname):
    msg = isctest.query.create(qname, "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.noerror(res2)
    isctest.check.adflag(res2)


def test_validation_recovery(ns2, ns4):
    # check recovery from spoofed server address.
    # prime cache with spoofed address records...
    msg = isctest.query.create("target.peer-ns-spoof", "A", cd=True)
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.servfail(res)
    ns4.rndc("dumpdb")
    with isctest.log.WatchLogFromStart("ns4/named_dump.db") as watcher:
        watcher.wait_for_line("10.53.0.100")

    # then reload server with properly signed zone
    shutil.copyfile(
        "ns2/peer.peer-ns-spoof.db.next", "ns2/peer.peer-ns-spoof.db.signed"
    )
    with ns2.watch_log_from_here() as watcher:
        ns2.rndc("reload peer.peer-ns-spoof")
        watcher.wait_for_line("zone peer.peer-ns-spoof/IN: loaded serial 2000042408")

    # and check we can resolve with the correct server address
    msg = isctest.query.create("test.target.peer-ns-spoof", "TXT")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.nxdomain(res)
    isctest.check.adflag(res)

    # check recovery from stripped DNSKEY RRSIG.
    # prime cache with spoofed address records...
    msg = isctest.query.create("dnskey-rrsigs-stripped", "DNSKEY", cd=True)
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.rr_count_eq(res.answer, 2)

    # then reload server with properly signed zone
    shutil.copyfile(
        "ns2/dnskey-rrsigs-stripped.db.next", "ns2/dnskey-rrsigs-stripped.db.signed"
    )
    with ns2.watch_log_from_here() as watcher:
        ns2.rndc("reload dnskey-rrsigs-stripped")
        watcher.wait_for_line(
            "zone dnskey-rrsigs-stripped/IN: loaded serial 2000042408"
        )

    # and check we can now resolve with the correct server address
    msg = isctest.query.create("b.dnskey-rrsigs-stripped", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)

    # check recovery from stripped DS RRSIG.
    # prime cache with spoofed address records...
    msg = isctest.query.create("child.ds-rrsigs-stripped", "DS", cd=True)
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.rr_count_eq(res.answer, 1)

    # then reload server with properly signed zone
    shutil.copyfile(
        "ns2/ds-rrsigs-stripped.db.next", "ns2/ds-rrsigs-stripped.db.signed"
    )
    with ns2.watch_log_from_here() as watcher:
        ns2.rndc("reload ds-rrsigs-stripped")
        watcher.wait_for_line("zone ds-rrsigs-stripped/IN: loaded serial 2000042408")

    # and check we can now resolve with the correct server address
    msg = isctest.query.create("b.child.ds-rrsigs-stripped", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.2")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)

    # check recovery with mismatching NS
    ns4.rndc("flush")
    msg = isctest.query.create("inconsistent", "NS", dnssec=False, cd=True)
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noadflag(res)
    isctest.check.rr_count_eq(res.answer, 2)
    isctest.check.rr_count_eq(res.additional, 0)

    msg = isctest.query.create("inconsistent", "NS", cd=True)
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noadflag(res)
    isctest.check.rr_count_eq(res.answer, 3)
    isctest.check.rr_count_eq(res.additional, 0)

    msg = isctest.query.create("inconsistent", "NS")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.adflag(res)
    isctest.check.rr_count_eq(res.answer, 3)
    isctest.check.rr_count_eq(res.additional, 0)


def test_failed_validation():
    # bogus zone
    msg = isctest.query.create("a.bogus.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.servfail(res)

    # missing key record
    msg = isctest.query.create("a.b.keyless.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.servfail(res)


def test_revoked_key():
    # validation should succeed if a revoked key is encountered
    msg = isctest.query.create("revkey.example", "SOA")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)


def test_standby_key():
    # check that a secure chain with one active and one inactive KSK
    # validates as secure
    msg = isctest.query.create("a.lazy-ksk", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)


def test_transitions():
    # check that a zone finishing transitioning from one algorithm
    # to another validates secure
    msg = isctest.query.create("algroll", "NS")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)

    # check that validation yields insecure during transition to signed
    msg = isctest.query.create("inprogress", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)
    a, _ = res.answer
    assert str(a[0]) == "10.53.0.10"


def test_validating_forwarder(ns4, ns9):
    # check validating forwarder behavior with mismatching NS
    ns4.rndc("flush")
    msg = isctest.query.create("inconsistent", "NS", dnssec=False, cd=True)
    res = isctest.query.tcp(msg, "10.53.0.9")
    isctest.check.noerror(res)
    isctest.check.rr_count_eq(res.answer, 2)
    isctest.check.rr_count_eq(res.additional, 0)
    isctest.check.noadflag(res)

    msg = isctest.query.create("inconsistent", "NS", cd=True)
    res = isctest.query.tcp(msg, "10.53.0.9")
    isctest.check.rr_count_eq(res.additional, 0)
    isctest.check.noadflag(res)
    isctest.check.rr_count_eq(res.answer, 3)
    isctest.check.rr_count_eq(res.authority, 0)
    isctest.check.rr_count_eq(res.additional, 0)

    msg.flags &= ~flags.CD
    res = isctest.query.tcp(msg, "10.53.0.9")
    isctest.check.rr_count_eq(res.answer, 3)
    isctest.check.rr_count_eq(res.authority, 0)
    isctest.check.rr_count_eq(res.additional, 0)
    isctest.check.adflag(res)

    # check validating forwarder sends CD to validate with a local trust anchor
    ns4.rndc("flush")
    msg = isctest.query.create("localkey.example", "SOA")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.servfail(res)

    with ns9.watch_log_from_here() as watcher:
        res = isctest.query.tcp(msg, "10.53.0.9")
        isctest.check.noerror(res)
        isctest.check.adflag(res)
        watcher.wait_for_line("status: SERVFAIL")


@pytest.mark.parametrize(
    "zone",
    [
        "missing-dnskey",
        "wrong-dnskey",
        "missing-ksk",
        "a.extradsunknownoid.example",
    ],
)
def test_missing_dnskey(zone, ns4):
    msg = isctest.query.create(f"a.{zone}", "A")
    res = isctest.query.tcp(msg, ns4.ip)
    isctest.check.servfail(res)
    isctest.check.ede(res, EDECode.DNSKEY_MISSING)


def test_expired_signatures(ns4):
    # check expired signatures do not validate
    msg = isctest.query.create("expired.example", "SOA")
    res = isctest.query.tcp(msg, "10.53.0.3")
    rrsig = res.get_rrset(
        res.answer,
        name.from_text("expired.example."),
        rdataclass.IN,
        rdatatype.RRSIG,
        rdatatype.SOA,
    )
    assert rrsig, "expected RRSIG(SOA) missing from AUTHORITY: " + str(rrsig)
    isctest.check.rr_count_eq(res.answer, 2)
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.servfail(res)
    isctest.check.noadflag(res)
    isctest.check.ede(res, EDECode.SIGNATURE_EXPIRED)
    assert Re("expired.example/.*: RRSIG has expired") in ns4.log

    # check future signatures do not validate
    msg = isctest.query.create("future.example", "SOA")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.servfail(res)
    isctest.check.noadflag(res)
    isctest.check.ede(res, EDECode.SIGNATURE_NOT_YET_VALID)
    assert Re("future.example/.*: RRSIG validity period has not begun") in ns4.log

    # check that a dynamic zone with future signatures is re-signed on load
    msg = isctest.query.create("managed-future.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.adflag(res)
    isctest.check.noerror(res)

    # test TTL is capped at RRSIG expiry time
    ns4.rndc("flush")
    msg = isctest.query.create("expiring.example", "SOA", cd=True)
    res1 = isctest.query.tcp(msg, "10.53.0.4")
    msg = isctest.query.create("expiring.example", "SOA")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res1)
    isctest.check.has_answer(res1)
    isctest.check.noerror(res2)
    isctest.check.has_answer(res2)
    for rrset in res1.answer:
        assert 240 <= rrset.ttl <= 300
    for rrset in res2.answer:
        assert rrset.ttl <= 60


def test_casing():
    # test legacy upper-case signer name validation
    msg = isctest.query.create("upper.example", "SOA")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.adflag(res)
    _, sig = res.answer
    assert sig.rdtype == rdatatype.RRSIG
    assert sig.covers == rdatatype.SOA
    assert "UPPER.EXAMPLE." in str(sig[0])

    # test that we lower-case signer name
    msg = isctest.query.create("LOWER.EXAMPLE", "SOA")
    res = isctest.query.tcp(msg, "10.53.0.4")
    _, sig = res.answer
    assert sig.rdtype == rdatatype.RRSIG
    assert sig.covers == rdatatype.SOA
    assert "lower.example." in str(sig[0])


def test_broken_servers():
    # check that a non-cacheable NODATA works
    msg = isctest.query.create("a.nosoa.secure.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.6")
    isctest.check.rr_count_eq(res1.authority, 0)
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res2)

    # check that a non-cacheable NXDOMAIN works
    msg = isctest.query.create("b.nosoa.secure.example", "TXT")
    res1 = isctest.query.tcp(msg, "10.53.0.6")
    isctest.check.rr_count_eq(res1.authority, 0)
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.nxdomain(res2)

    # check that split RRSIGs are handled
    msg = isctest.query.create("split-rrsig", "SOA")
    res = isctest.query.tcp(msg, "10.53.0.6")
    soa, _ = res.answer
    assert soa[0].serial > 1

    # check that not-at-zone-apex RRSIG(SOA) rrsets are removed
    msg = isctest.query.create("split-rrsig", "AXFR")
    res = isctest.query.tcp(msg, "10.53.0.6")

    nza = [
        r
        for r in res.answer
        if str(r.name) == "not-at-zone-apex.split-rrsig."
        and r.rdtype == rdatatype.RRSIG
        and r.covers == rdatatype.SOA
    ]
    assert not nza

    # check validation with missing nearest encloser proof
    msg = isctest.query.create("b.c.d.optout-tld", "DS")
    res = isctest.query.tcp(msg, "10.53.0.6")
    nsec3s = [a for a in res.authority if a.rdtype == rdatatype.NSEC3]
    assert len(nsec3s) == 2

    msg = isctest.query.create("b.c.d.optout-tld", "A")
    res = isctest.query.tcp(msg, "10.53.0.6")
    nsec3s = [a for a in res.authority if a.rdtype == rdatatype.NSEC3]
    assert len(nsec3s) == 1

    res = isctest.query.tcp(msg, "10.53.0.6")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)
    msg = isctest.query.create("optout-tld", "SOA")
    res = isctest.query.tcp(msg, "10.53.0.6")
    isctest.check.noadflag(res)


def test_pending_ds(ns4):
    # check that a query against a validating resolver succeeds when there is
    # a negative cache entry with trust level "pending" for the DS.  prime
    # with a +cd DS query to produce the negative cache entry, then send a
    # query that uses that entry as part of the validation process.
    ns4.rndc("flush")
    msg = isctest.query.create("insecure.example", "DS", cd=True)
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.rr_count_eq(res.answer, 0)
    isctest.check.rr_count_eq(res.authority, 4)
    msg = isctest.query.create("a.insecure.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.rr_count_eq(res.answer, 1)
    isctest.check.rr_count_eq(res.authority, 0)
    isctest.check.noadflag(res)


def test_unknown_algorithms():
    # check that unknown DNSKEY algorithm validates as insecure
    msg = isctest.query.create("dnskey-unknown.example", "A", dnssec=False)
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res1)
    isctest.check.noerror(res2)
    isctest.check.noadflag(res2)

    # check that unsupported DNSKEY algorithms are in the DNSKEY RRsets
    msg = isctest.query.create("dnskey-unsupported.example", "DNSKEY")
    res = isctest.query.tcp(msg, "10.53.0.3")
    isctest.check.noerror(res)

    msg = isctest.query.create("dnskey-unsupported-2.example", "DNSKEY")
    res = isctest.query.tcp(msg, "10.53.0.3")
    isctest.check.noerror(res)
    rrsets = [str(r) for r in res.answer]
    assert any("257 3 255" in r for r in rrsets)

    # check that unsupported DNSKEY algorithm validates as insecure
    msg = isctest.query.create("dnskey-unsupported.example", "DNSKEY")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)
    isctest.check.ede(res, EDECode.UNSUPPORTED_DNSKEY_ALGORITHM)

    # check that DNSKEY with an unsupported reserve key validates
    msg = isctest.query.create("dnskey-unsupported-2.example", "DNSKEY")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.adflag(res)

    # check EDE code 2 for unsupported DS digest algorithm
    msg = isctest.query.create("a.ds-unsupported.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.ede(res, EDECode.UNSUPPORTED_DS_DIGEST_TYPE)

    # check EDE code 1 for bad algorithm mnemonic
    msg = isctest.query.create("badalg.secure.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.noadflag(res)
    isctest.check.ede(res, EDECode.UNSUPPORTED_DNSKEY_ALGORITHM)

    # check that zone contents are still secure despite disable-algorithms
    # on query name (name below zone name).
    msg = isctest.query.create("z.secure.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.rr_count_eq(res.answer, 2)
    isctest.check.noerror(res)
    isctest.check.adflag(res)

    # check that zone contents are trated insecure (name above zone name).
    msg = isctest.query.create("zonecut.ent.secure.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.rr_count_eq(res.answer, 2)
    isctest.check.noerror(res)
    isctest.check.noadflag(res)
    isctest.check.ede(res, EDECode.UNSUPPORTED_DNSKEY_ALGORITHM)

    # check that DS records are still treated as secure at the
    # disable-algorithm name
    msg = isctest.query.create("badalg.secure.example", "DS")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.rr_count_eq(res.answer, 2)
    isctest.check.noerror(res)
    isctest.check.adflag(res)

    # check both EDE code 1 and 2 for unsupported digest on one DNSKEY
    # and unsupported algorithm on the other
    msg = isctest.query.create("a.digest-alg-unsupported.example", "A")
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noadflag(res)
    isctest.check.ede(res, EDECode.UNSUPPORTED_DNSKEY_ALGORITHM)
    isctest.check.ede(res, EDECode.UNSUPPORTED_DS_DIGEST_TYPE)

    # check that unknown DNSKEY algorithm + unknown NSEC3 hash algorithm
    # validates as insecure
    msg = isctest.query.create("dnskey-nsec3-unknown.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res1)
    isctest.check.noerror(res2)
    isctest.check.noadflag(res2)


def test_legacy_dnssec_types_are_signed():
    """SIG (24), NXT (30) and KEY (25) records carry a covering RRSIG.

    Per RFC 3755 SIG and NXT are obsolete and treated as opaque zone
    data; KEY remains valid for SIG(0)/TKEY use.  All three are
    ordinary zone data and must be signed like any other RRset.
    """
    for owner, rrtype in [
        ("sigrr.secure.example", "SIG"),
        ("nxtrr.secure.example", "NXT"),
        ("keyrr.secure.example", "KEY"),
    ]:
        expected = rdatatype.from_text(rrtype)
        msg = isctest.query.create(owner, rrtype, cd=True)
        res = isctest.query.tcp(msg, "10.53.0.3")
        isctest.check.noerror(res)
        assert any(
            rr.rdtype == expected for rr in res.answer
        ), f"{rrtype} record missing in answer for {owner}: {res.answer}"
        assert any(
            rr.rdtype == rdatatype.RRSIG and rr.covers == expected for rr in res.answer
        ), f"RRSIG({rrtype}) missing in answer for {owner}: {res.answer}"


def test_rrsigs_for_glue():
    msg = isctest.query.create("ns3.secure.example", "A", cd=True)
    res = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noerror(res)
    isctest.check.rr_count_eq(res.answer, 2)
    assert any(record.rdtype == rdatatype.A for record in res.answer)
    assert any(
        record.rdtype == rdatatype.RRSIG and record.covers == rdatatype.A
        for record in res.answer
    )


def test_extra_bad_algorithm():
    msg = isctest.query.create("a.extrabadkey.example", "A")
    res1 = isctest.query.tcp(msg, "10.53.0.3")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.same_answer(res1, res2)
    isctest.check.adflag(res2)



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os

import isctest
import isctest.mark

pytestmark = [isctest.mark.with_dnstap]


def line_to_query(line):
    # dnstap-read output line example
    # 05-Feb-2026 11:00:57.853 RQ 10.53.0.6:38507 -> 10.53.0.3:22047 TCP 56b fooXXX.example./IN/NS
    _, _, _, _, _, _, _, _, query = line.split(" ", 9)
    return query


def extract_dnstap(ns):
    ns.rndc("dnstap -roll 1")
    path = os.path.join(ns.identifier, "dnstap.out.0")
    dnstapread = isctest.run.cmd(
        [isctest.vars.ALL["DNSTAPREAD"], path],
    )

    lines = dnstapread.out.splitlines()
    return list(map(line_to_query, lines))


def test_auth_res_deleg(ns2):
    msg = isctest.query.create("aaaa.sub.example.", "AAAA")
    res = isctest.query.udp(msg, ns2.ip)
    isctest.check.noerror(res)
    assert len(res.answer[0]) == 1
    res.answer[0].ttl = 300
    assert str(res.answer[0]) == "aaaa.sub.example. 300 IN AAAA ac::dc"

    queries = extract_dnstap(ns2)
    assert len(queries) == 1
    assert queries[0] == "aaaa.sub.example/IN/AAAA"



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from collections.abc import AsyncGenerator

import dns.edns
import dns.name
import dns.rcode
import dns.rdataclass
import dns.rdatatype
import dns.rrset

from isctest.asyncserver import (
    AsyncDnsServer,
    DnsResponseSend,
    DomainHandler,
    QnameHandler,
    QnameQtypeHandler,
    QueryContext,
    StaticResponseHandler,
)


def rrset(
    qname: dns.name.Name | str,
    rtype: dns.rdatatype.RdataType,
    rdata: str,
    ttl: int = 300,
) -> dns.rrset.RRset:
    return dns.rrset.from_text(qname, ttl, dns.rdataclass.IN, rtype, rdata)


class RootNsHandler(QnameQtypeHandler, StaticResponseHandler):
    qnames = ["."]
    qtypes = [dns.rdatatype.NS]
    answer = [rrset(".", dns.rdatatype.NS, "a.root-servers.nil.")]
    additional = [rrset("a.root-servers.nil.", dns.rdatatype.A, "10.53.0.3")]


class ExampleNsHandler(QnameQtypeHandler, StaticResponseHandler):
    qnames = ["example."]
    qtypes = [dns.rdatatype.NS]
    answer = [rrset("example.", dns.rdatatype.NS, "ns.example.")]
    additional = [rrset("ns.example.", dns.rdatatype.A, "10.53.0.3")]


class ExampleCookieHandler(DomainHandler):
    domains = ["example."]

    def _get_cookie(self, qctx: QueryContext) -> dns.edns.CookieOption | None:
        for o in qctx.query.options:
            if o.otype == dns.edns.OptionType.COOKIE:
                cookie = o
                cookie.server = b"\x11\x22\x33\x44\x55\x66\x77\x88"
                return cookie

        return None

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        if cookie := self._get_cookie(qctx):
            # If there is a client cookie, mock BADCOOKIE to trigger
            # the resend loop logic.
            qctx.response.use_edns(options=[cookie])
            qctx.response.set_rcode(dns.rcode.BADCOOKIE)
            yield DnsResponseSend(qctx.response)
        else:
            # If missing cookie entirely, just return SERVFAIL
            qctx.response.set_rcode(dns.rcode.SERVFAIL)
            yield DnsResponseSend(qctx.response)


class TestDotComServFailHandler(QnameHandler, StaticResponseHandler):
    qnames = ["test.com."]
    authoritative = False
    rcode = dns.rcode.SERVFAIL


def main() -> None:
    server = AsyncDnsServer(default_aa=True, default_rcode=dns.rcode.NOERROR)
    server.install_response_handlers(
        RootNsHandler(),
        ExampleNsHandler(),
        ExampleCookieHandler(),
        TestDotComServFailHandler(),
    )
    server.run()


if __name__ == "__main__":
    main()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from pathlib import Path

import sys

SYSTEM_TEST_DIR_GIT_PATH = "bin/tests/system"


def load_vars_from_build_files() -> dict[str, str]:
    # TOP_BUILDDIR is special, it is always read from the source directory
    top_builddir_file = Path(__file__).resolve().parent / ".build_vars" / "TOP_BUILDDIR"
    if not top_builddir_file.exists():
        raise RuntimeError(
            'Uninitialized build variable: "TOP_BUILDDIR". Did you run `meson compile`?'
        )

    top_builddir = top_builddir_file.read_text(encoding="utf-8").strip()

    build_vars = {
        "TOP_BUILDDIR": top_builddir,
    }

    var_dir = (
        Path(top_builddir)
        / SYSTEM_TEST_DIR_GIT_PATH
        / "isctest"
        / "vars"
        / ".build_vars"
    )

    for var in [
        "CURL",
        "FSTRM_CAPTURE",
        "JQ",
        "NC",
        "PERL",
        "PYTEST",
        "PYTHON",
        "SHELL",
        "TOP_SRCDIR",
        "XMLLINT",
        "XSLTPROC",
    ]:
        var_file = var_dir / var
        if var_file.exists():
            build_vars[var] = var_file.read_text(encoding="utf-8").strip()

    # When running inside a virtualenv, prefer the virtualenv's interpreter
    # over the path baked in by meson.
    # This is needed because meson is configured to prefer specific versions
    # (e.g. python3.12 and python3.11) by default to properly detect the
    # right python interpreter in CI - this is undesirable when running
    # in a virtualenv as it can lead to the system tests using the system
    # versions of Python dependencies defeating the purpose of the virtual
    # environment.
    if sys.prefix != sys.base_prefix:
        build_vars["PYTHON"] = sys.executable

    return build_vars


BUILD_VARS = load_vars_from_build_files()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os

import requests

import isctest.mark

pytestmark = [isctest.mark.with_json_c, isctest.mark.with_developer]


def get_delegdb_watermarks(ip, port):
    watermarks = []
    r = requests.get(f"http://{ip}:{port}/json/v1/mem", timeout=600)
    assert r.status_code == 200
    mem = r.json()["memory"]
    for c in mem["contexts"]:
        if c["name"] == "dns_delegdb":
            watermarks.append((c["id"], c["lowater"], c["hiwater"]))
    return watermarks


def check_watermarks(watermarks1, watermarks2):
    if watermarks2 is not None:
        assert len(watermarks1) == len(watermarks2)
    for i, (id1, lowater1, hiwater1) in enumerate(watermarks1):
        assert lowater1 > 0
        assert hiwater1 > 0
        if watermarks2 is not None:
            id2, lowater2, hiwater2 = watermarks2[i]
            assert id1 != id2
            assert lowater1 == lowater2
            assert hiwater1 == hiwater2


def test_delegdb_flush(ns1):
    statsport = os.getenv("EXTRAPORT1")

    watermarks1 = get_delegdb_watermarks(ns1.ip, statsport)
    check_watermarks(watermarks1, None)

    with ns1.watch_log_from_here() as watcher:
        ns1.rndc("flush")
        watcher.wait_for_sequence(
            ["flushing caches in all views succeeded", "loop exclusive mode: ended"]
        )

    # The previous delegdb contexts can still be hanging around for a little
    # bit, until RCU reclamation run and it actually gets detached/freed.
    for watermarks in watermarks1:
        id1, _, _ = watermarks
        with ns1.watch_log_from_start() as watcher:
            watcher.wait_for_line(f"destroyed mctx {id1}")

    watermarks2 = get_delegdb_watermarks(ns1.ip, statsport)
    check_watermarks(watermarks1, watermarks2)



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import glob
import os
import subprocess

import dns.rcode
import pytest

import isctest

pytestmark = pytest.mark.extra_artifacts(["conf/*.conf"])


def test_hooks():
    msg = isctest.query.create("example.com.", "A")
    res = isctest.query.udp(msg, "10.53.0.1")
    # the test-async plugin changes the status of any positive answer to NOTIMP
    isctest.check.notimp(res)


def test_hooks_noextension(ns1, templates):
    templates.render("ns1/named.conf", {"noextension": True})
    with ns1.watch_log_from_here() as watcher:
        ns1.rndc("reload")
        watcher.wait_for_line("all zones loaded")
    test_hooks()


def test_hooks_global_hook():
    msg = isctest.query.create("idontexists.example.com", "A")
    res = isctest.query.udp(msg, "10.53.0.2")
    isctest.check.rcode(res, dns.rcode.NOERROR)


def test_hooks_zone_hook1():
    msg = isctest.query.create("idonotexists.example2.com", "A")
    res = isctest.query.udp(msg, "10.53.0.2")
    isctest.check.rcode(res, dns.rcode.SERVFAIL)


def test_hooks_zone_hook2():
    msg = isctest.query.create("idonotexists.example3.com", "A")
    res = isctest.query.udp(msg, "10.53.0.2")
    isctest.check.rcode(res, dns.rcode.NOTIMP)


# ensure a plugin defined in a template is correcty registered to zone using
# this template
def test_hooks_zonetemplate1():
    msg = isctest.query.create("idonotexists.example4.com", "A")
    res = isctest.query.udp(msg, "10.53.0.2")
    isctest.check.rcode(res, dns.rcode.NOTAUTH)


# ensure plugin defined in zone are correctly registered when the zone also
# using a template with plugins (the plugin defined in the template is called
# first, but it bails out without doing anything because the first label is
# "skipfoo". So the plugin defined in the zone is then called, and return
# notzone)
def test_hooks_zonetemplate2():
    msg = isctest.query.create("skipfoo.example4.com", "A")
    res = isctest.query.udp(msg, "10.53.0.2")
    isctest.check.rcode(res, dns.rcode.NOTZONE)

    # test that the skip label is not matched add suffix
    msg = isctest.query.create("skipfooX.example4.com", "A")
    res = isctest.query.udp(msg, "10.53.0.2")
    isctest.check.rcode(res, dns.rcode.NOTAUTH)

    # test that the skip label is not matched add prefix
    msg = isctest.query.create("Xskipfoo.example4.com", "A")
    res = isctest.query.udp(msg, "10.53.0.2")
    isctest.check.rcode(res, dns.rcode.NOTAUTH)


def test_hooks_zone_rndc_reload(servers):
    ns2 = servers["ns2"]
    ns2.rndc("reload")


def test_hooks_checkconf():
    for filename in glob.glob("conf/good*.conf"):
        isctest.run.cmd([os.environ["CHECKCONF"], filename])
    for filename in glob.glob("conf/bad*.conf"):
        with pytest.raises(subprocess.CalledProcessError):
            isctest.run.cmd([os.environ["CHECKCONF"], filename])



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os

from .build import BUILD_VARS  # type: ignore

BASIC_VARS = {
    "ARPANAME": f"{BUILD_VARS['TOP_BUILDDIR']}/arpaname",
    "CDS": f"{BUILD_VARS['TOP_BUILDDIR']}/dnssec-cds",
    "CHECKCONF": f"{BUILD_VARS['TOP_BUILDDIR']}/named-checkconf",
    "CHECKZONE": f"{BUILD_VARS['TOP_BUILDDIR']}/named-checkzone",
    "DELV": f"{BUILD_VARS['TOP_BUILDDIR']}/delv",
    "DIG": f"{BUILD_VARS['TOP_BUILDDIR']}/dig",
    "DNSTAPREAD": f"{BUILD_VARS['TOP_BUILDDIR']}/dnstap-read",
    "DSFROMKEY": f"{BUILD_VARS['TOP_BUILDDIR']}/dnssec-dsfromkey",
    "FEATURETEST": f"{BUILD_VARS['TOP_BUILDDIR']}/feature-test",
    "HOST": f"{BUILD_VARS['TOP_BUILDDIR']}/host",
    "IMPORTKEY": f"{BUILD_VARS['TOP_BUILDDIR']}/dnssec-importkey",
    "JOURNALPRINT": f"{BUILD_VARS['TOP_BUILDDIR']}/named-journalprint",
    "KEYFRLAB": f"{BUILD_VARS['TOP_BUILDDIR']}/dnssec-keyfromlabel",
    "KEYGEN": f"{BUILD_VARS['TOP_BUILDDIR']}/dnssec-keygen",
    "KSR": f"{BUILD_VARS['TOP_BUILDDIR']}/dnssec-ksr",
    "MAKEJOURNAL": f"{BUILD_VARS['TOP_BUILDDIR']}/named-makejournal",
    "MDIG": f"{BUILD_VARS['TOP_BUILDDIR']}/mdig",
    "NAMED": f"{BUILD_VARS['TOP_BUILDDIR']}/named",
    "NSEC3HASH": f"{BUILD_VARS['TOP_BUILDDIR']}/nsec3hash",
    "NSLOOKUP": f"{BUILD_VARS['TOP_BUILDDIR']}/nslookup",
    "NSUPDATE": f"{BUILD_VARS['TOP_BUILDDIR']}/nsupdate",
    "NZD2NZF": f"{BUILD_VARS['TOP_BUILDDIR']}/named-nzd2nzf",
    "REVOKE": f"{BUILD_VARS['TOP_BUILDDIR']}/dnssec-revoke",
    "RNDC": f"{BUILD_VARS['TOP_BUILDDIR']}/rndc",
    "RNDCCONFGEN": f"{BUILD_VARS['TOP_BUILDDIR']}/rndc-confgen",
    "RRCHECKER": f"{BUILD_VARS['TOP_BUILDDIR']}/named-rrchecker",
    "SETTIME": f"{BUILD_VARS['TOP_BUILDDIR']}/dnssec-settime",
    "SIGNER": f"{BUILD_VARS['TOP_BUILDDIR']}/dnssec-signzone",
    "TSIGKEYGEN": f"{BUILD_VARS['TOP_BUILDDIR']}/tsig-keygen",
    "VERIFY": f"{BUILD_VARS['TOP_BUILDDIR']}/dnssec-verify",
    "WIREFORMAT": f"{BUILD_VARS['TOP_BUILDDIR']}/named-wireformat",
    "BIGKEY": f"{BUILD_VARS['TOP_BUILDDIR']}/bigkey",
    "PIPEQUERIES": f"{BUILD_VARS['TOP_BUILDDIR']}/pipequeries",
    "TMPDIR": os.getenv("TMPDIR", "/tmp"),
    "KRB5_CONFIG": "/dev/null",  # we don't want a KRB5_CONFIG setting breaking the tests
    "LC_ALL": "C",
    "ANS_LOG_LEVEL": "debug",
    "DYLIB": ("dylib" if os.uname().sysname == "Darwin" else "so"),
}



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from pathlib import Path
from re import compile as Re

import os
import re
import time

from cryptography.hazmat.primitives.asymmetric import rsa
from dns.dnssectypes import NSEC3Hash
from dns.rdtypes.dnskeybase import Flag

import dns.dnssec
import dns.name
import dns.rdata
import dns.rdataclass
import dns.rdatatype
import dns.zone
import pytest

from isctest.template import NO_NS, zones
from isctest.zone import Zone

import isctest

pytestmark = pytest.mark.extra_artifacts(
    [
        "verify.out.*",
        "zones/K*",
        "zones/dsset-*",
        "zones/*.bad",
        "zones/*.good",
        "zones/*.out*",
        "zones/*.tmp",
        "zones/updated*",
        "zones/bad-nsec3param-hash*",
        "zones/nsec-bad-nsec3param-hash*",
        "zones/good-bad-nsec3param-hash*",
        "zones/nsec+non-zero-nsec3param-flags*",
        "zones/no-nsec+non-zero-nsec3param-flags*",
        "zones/nseconly-nsec3param*",
    ]
)

VERIFY = os.environ.get("VERIFY")


def bootstrap():

    def generate_keys():
        algorithm = dns.dnssec.Algorithm.RSASHA256
        ksk_private_key = rsa.generate_private_key(public_exponent=65537, key_size=1024)
        try:
            ksk_dnskey = dns.dnssec.make_dnskey(
                public_key=ksk_private_key.public_key(),
                algorithm=algorithm,
                flags=Flag.ZONE | Flag.SEP,
            )
        except ImportError as exc:
            # if the cryptography package is too old, the make_dnskey() function
            # will raise ImportError at runtime
            pytest.skip(f"{exc}")
        return ksk_private_key, ksk_dnskey

    def gen_and_sign(name, nsec3=False):
        zone = Zone(name, NO_NS, signed=True)
        ksk_private_key, ksk_dnskey = generate_keys()
        keys = [(ksk_private_key, ksk_dnskey)]
        zone.render()

        # read the rendered zone
        unsigned_path = str(Path(zone.ns.name) / zone.filepath_unsigned)
        signed_path = str(Path(zone.ns.name) / zone.filepath_signed)
        zoneobj = dns.zone.from_file(unsigned_path, origin=f"{name}.")
        lifetime = 30 * 86400

        # sign the zone
        with zoneobj.writer() as txn:
            dns.dnssec.sign_zone(
                zone=zoneobj,
                txn=txn,
                keys=keys,
                lifetime=lifetime,
                add_dnskey=True,
                deterministic=False,  # for OpenSSL<3.2.0 compat
            )

        # This generates an NSEC3 record for the apex, so that we can
        # verify a zone with both a valid and invalid NSEC3PARAM record.
        if nsec3:
            origin = dns.name.from_text(f"{name}.")
            hashname = dns.dnssec.nsec3_hash(
                origin, salt=b"", iterations=0, algorithm=NSEC3Hash.SHA1
            )
            nsec3rdata = dns.rdata.from_text(
                rdclass=dns.rdataclass.IN,
                rdtype=dns.rdatatype.NSEC3,
                tok=f"1 0 0 - {hashname} SOA NS A NSEC3PARAM DNSKEY RRSIG",
            )
            nsec3_owner = dns.name.from_text(f"{hashname}.{name}.")
            node = zoneobj.find_node(nsec3_owner, create=True)
            nsec3_rrset = node.find_rdataset(
                rdclass=dns.rdataclass.IN, rdtype=dns.rdatatype.NSEC3, create=True
            )
            nsec3_rrset.ttl = 300
            nsec3_rrset.add(nsec3rdata)

            inception = int(time.time()) - 30
            nsec3_sigdata = dns.dnssec.sign(
                rrset=(nsec3_owner, nsec3_rrset),
                private_key=ksk_private_key,
                signer=origin,
                dnskey=ksk_dnskey,
                inception=inception,
                lifetime=lifetime,
            )
            nsec3_rrsig = node.find_rdataset(
                rdclass=dns.rdataclass.IN,
                rdtype=dns.rdatatype.RRSIG,
                covers=dns.rdatatype.NSEC3,
                create=True,
            )
            nsec3_rrsig.ttl = 300
            nsec3_rrsig.add(nsec3_sigdata)

        zoneobj.to_file(signed_path)

        return zone

    return {
        "zones": zones(
            [
                gen_and_sign("bad-nsec3param-hash"),
                gen_and_sign("nsec-bad-nsec3param-hash"),
                gen_and_sign("good-bad-nsec3param-hash", True),
                gen_and_sign("nsec+non-zero-nsec3param-flags"),
                gen_and_sign("no-nsec+non-zero-nsec3param-flags"),
                gen_and_sign("nseconly-nsec3param", True),
            ]
        ),
    }


@pytest.mark.parametrize(
    "zone",
    [
        "ksk-only.nsec3",
        "ksk-only.nsec",
        "ksk+zsk.nsec3.apex-dname",
        "ksk+zsk.nsec3",
        "ksk+zsk.nsec.apex-dname",
        "ksk+zsk.nsec",
        "ksk+zsk.optout",
        "zsk-only.nsec3",
        "zsk-only.nsec",
    ],
)
def test_verify_good_zone_files(zone):
    isctest.run.cmd([VERIFY, "-z", "-o", zone, f"zones/{zone}.good"])


def test_verify_good_zone_nsec_next_name_case_mismatch():
    isctest.run.cmd(
        [
            VERIFY,
            "-o",
            "nsec-next-name-case-mismatch",
            "zones/nsec-next-name-case-mismatch.good",
        ],
    )


def verify_bad_zone(zone):
    only_opt = ["-z"] if re.search(r"^[zk]sk-only", zone) else []
    cmd = isctest.run.cmd(
        [VERIFY, *only_opt, "-o", zone, f"zones/{zone}.bad"],
        raise_on_exception=False,
    )
    assert cmd.rc != 0
    return cmd


@pytest.mark.parametrize(
    "zone",
    [
        "ksk-only.dnskeyonly",
        "ksk+zsk.dnskeyonly",
        "zsk-only.dnskeyonly",
    ],
)
def test_verify_bad_zone_files_dnskeyonly(zone):
    cmd = verify_bad_zone(zone)
    assert "DNSKEY is not signed" in cmd.err


@pytest.mark.parametrize(
    "zone",
    [
        "ksk-only.nsec3.expired",
        "ksk-only.nsec.expired",
        "ksk+zsk.nsec3.expired",
        "ksk+zsk.nsec.expired",
        "ksk+zsk.nsec.ksk-expired",
        "zsk-only.nsec3.expired",
        "zsk-only.nsec.expired",
        "ksk+zsk.nsec3.ksk-expired",
    ],
)
def test_verify_bad_zone_files_expired(zone):
    cmd = verify_bad_zone(zone)
    assert Re("signature has expired|No self-signed DNSKEY found") in cmd.err


@pytest.mark.parametrize(
    "zone",
    [
        "ksk+zsk.nsec.out-of-zone-nsec",
        "ksk+zsk.nsec.below-bottom-of-zone-nsec",
        "ksk+zsk.nsec.below-dname-nsec",
    ],
)
def test_verify_bad_zone_files_unexpected_nsec_rrset(zone):
    cmd = verify_bad_zone(zone)
    assert "unexpected NSEC RRset at" in cmd.err


def test_verify_bad_zone_files_bad_nsec_record():
    cmd = verify_bad_zone("ksk+zsk.nsec.broken-chain")
    assert Re("Bad NSEC record for.*, next name mismatch") in cmd.err


def test_verify_bad_zone_files_bad_bitmap():
    cmd = verify_bad_zone("ksk+zsk.nsec.bad-bitmap")
    assert "bit map mismatch" in cmd.err


def test_verify_bad_zone_files_missing_nsec3_record():
    cmd = verify_bad_zone("ksk+zsk.nsec3.missing-empty")
    assert "Missing NSEC3 record for" in cmd.err


def test_verify_bad_zone_files_no_dnssec_keys():
    cmd = verify_bad_zone("unsigned")
    assert "Zone contains no DNSSEC keys" in cmd.err


def test_verify_bad_zone_files_unequal_nsec3_chains():
    cmd = verify_bad_zone("ksk+zsk.nsec3.extra-nsec3")
    assert "Expected and found NSEC3 chains not equal" in cmd.err


# checking error message when -o is not used
# and a SOA record not at top of zone is found
def test_verify_soa_not_at_top_error():
    # when -o is not used, origin is set to zone file name,
    # which should cause an error in this case
    cmd = isctest.run.cmd([VERIFY, "zones/ksk+zsk.nsec.good"], raise_on_exception=False)
    assert "not at top of zone" in cmd.err
    assert "use -o to specify a different zone origin" in cmd.err


# checking error message when an invalid -o is specified
# and a SOA record not at top of zone is found
def test_verify_invalid_o_option_soa_not_at_top_error():
    cmd = isctest.run.cmd(
        [VERIFY, "-o", "invalid.origin", "zones/ksk+zsk.nsec.good"],
        raise_on_exception=False,
    )
    assert "not at top of zone" in cmd.err
    assert "use -o to specify a different zone origin" not in cmd.err


# checking dnssec-verify -J reads journal file
def test_verify_j_reads_journal_file():
    cmd = isctest.run.cmd(
        [
            VERIFY,
            "-o",
            "updated",
            "-J",
            "zones/updated.other.jnl",
            "zones/updated.other",
        ]
    )
    assert "Loading zone 'updated' from file 'zones/updated.other'" in cmd.out


# checking that unknown hash is detected
def test_verify_rejects_bad_nsec3param_hash():
    cmd = isctest.run.cmd(
        [
            VERIFY,
            "-z",
            "-o",
            "bad-nsec3param-hash",
            "zones/bad-nsec3param-hash.db.signed",
        ],
        raise_on_exception=False,
    )
    assert cmd.rc != 0
    assert "No usable NSEC/NSEC3 chain for testing" in cmd.err


def test_verify_ignores_bad_nsec3param_hash_with_nsec():
    isctest.run.cmd(
        [
            VERIFY,
            "-z",
            "-o",
            "nsec-bad-nsec3param-hash",
            "zones/nsec-bad-nsec3param-hash.db.signed",
        ]
    )


def test_verify_ignores_bad_nsec3param_hash_with_good_nsec3param():
    isctest.run.cmd(
        [
            VERIFY,
            "-z",
            "-o",
            "good-bad-nsec3param-hash",
            "zones/good-bad-nsec3param-hash.db.signed",
        ]
    )


def test_verify_rejects_no_nsec_and_all_nsec3param_with_non_zero_flags():
    cmd = isctest.run.cmd(
        [
            VERIFY,
            "-z",
            "-o",
            "no-nsec+non-zero-nsec3param-flags",
            "zones/no-nsec+non-zero-nsec3param-flags.db.signed",
        ],
        raise_on_exception=False,
    )
    assert cmd.rc != 0
    assert "No usable NSEC/NSEC3 chain for testing" in cmd.err


def test_verify_ignores_nsec3param_non_zero_flags():
    isctest.run.cmd(
        [
            VERIFY,
            "-z",
            "-o",
            "nsec+non-zero-nsec3param-flags",
            "zones/nsec+non-zero-nsec3param-flags.db.signed",
        ],
    )


def test_verify_rejects_nsec3param_with_nsec_only_key():
    cmd = isctest.run.cmd(
        [
            VERIFY,
            "-z",
            "-o",
            "nseconly-nsec3param",
            "zones/nseconly-nsec3param.db.signed",
        ],
        raise_on_exception=False,
    )
    assert cmd.rc != 0
    isctest.log.debug(cmd.err)
    assert "cannot use NSEC3 with key algorithm" in cmd.out



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import requests

import isctest
import isctest.mark

pytestmark = [isctest.mark.live_internet_test]

# URL for the official root hints file
NAMED_ROOT_URL = "https://www.internic.net/zones/named.root"


def strip_comments(text):
    """Return only the resource records, without comments and blank lines."""
    records = []
    for line in text.splitlines():
        line = line.split(";", 1)[0].rstrip()
        if line:
            records.append(line)
    return records


def test_named_root_hints():
    """
    Test that 'named -H' output matches the official
    root hints from https://www.internic.net/zones/named.root
    """
    resp = requests.get(NAMED_ROOT_URL, timeout=30)
    resp.raise_for_status()

    internic_records = strip_comments(resp.text)

    named = isctest.vars.ALL["NAMED"]
    cmd = isctest.run.cmd([named, "-H"])
    builtin_records = strip_comments(cmd.out)

    assert (
        internic_records == builtin_records
    ), "Built-in root hints differ from official named.root"



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.


import dns.flags
import dns.message

import isctest


def test_gluecache_inzone_ns_target(ns1):
    """Exercise the glue cache path where the NS target is authoritative
    in-zone data (not glue below a zone cut).

    When sub.example. is delegated to ns.example. and ns.example. lives
    at the zone apex, qpzone_find returns ISC_R_SUCCESS rather than
    DNS_R_GLUE in glue_nsdname_cb.  The address records are not treated
    as glue and are therefore not included in the additional section of
    the referral.  However, the rdataset returned by qpzone_find is
    associated but never cleaned up, leaking a vecheader reference.

    Query multiple times to exercise both the cache-miss and cache-hit
    paths in glue_nsdname_cb, then shut down named and verify no
    memory was leaked.
    """
    for _ in range(3):
        msg = isctest.query.create("foo.sub.example.", "A")
        msg.flags &= ~dns.flags.RD
        res = isctest.query.udp(msg, "10.53.0.1")

        expected = """;ANSWER
;AUTHORITY
sub.example. 300 IN NS ns.example.
;ADDITIONAL
"""
        expected_msg = dns.message.from_text(expected)

        isctest.check.noerror(res)
        isctest.check.rrsets_equal(res.answer, expected_msg.answer)
        isctest.check.rrsets_equal(res.authority, expected_msg.authority)
        isctest.check.rrsets_equal(res.additional, expected_msg.additional)

    # Stop the server and check for memory leaks in the shutdown log.
    # The leaked vecheader reference will show up as outstanding memory
    # allocations from rdatavec.c in named's memory tracking output.
    ns1.stop()
    assert "outstanding memory" not in ns1.log



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from re import compile as Re

import time

import isctest

DUMP_FILE_NAME = "ns2/named_dump.db"


def warmup_cache(ns):
    # Let's flush all caches first so we're sure the query will fully run into the resolver.
    flush_caches(ns, "flush", "", "flushing caches in all views succeeded")
    msg = isctest.query.create("gee.foo.bar.com.", "A")
    res = isctest.query.udp(msg, ns.ip)
    isctest.check.noerror(res)
    msg = isctest.query.create("whatever.lame-and-expired-soon.bar.com.", "A")
    # This will fail, we don't care, this is just to have the entry in the cache
    isctest.query.udp(msg, ns.ip)
    assert len(res.answer) == 1


def check_cache_expired(ns):
    time.sleep(2)
    with ns.watch_log_from_here() as watcher:
        ns.rndc("dumpdb -expired")
        watcher.wait_for_line("dumpdb complete")
    with isctest.log.WatchLogFromStart(DUMP_FILE_NAME) as watcher:
        patterns = [
            Re("lame-and-expired-soon.bar.com. 0 DELEG server-name=ns.somehost.com."),
            Re("foo.bar.com. 99999[5-9] DELEG server-name=ns.somehost.com."),
        ]
        watcher.wait_for_all(patterns)


def check_cache(ns, hit):
    with ns.watch_log_from_here() as watcher:
        ns.rndc("dumpdb -deleg")
        watcher.wait_for_line("dumpdb complete")
    with isctest.log.WatchLogFromStart(DUMP_FILE_NAME) as watcher:
        if hit:
            pattern = Re("foo.bar.com. 99999[5-9] DELEG server-name=ns.somehost.com.")
            watcher.wait_for_line(pattern)
        else:
            seq = ["; Delegation cache", ";", ";", "; Start view _bind"]
            watcher.wait_for_sequence(seq)


def reload_server(ns):
    with ns.watch_log_from_here() as watcher:
        ns.rndc("reload")
        watcher.wait_for_line("running")
    with ns.watch_log_from_here() as watcher:
        ns.rndc("reconfig")
        watcher.wait_for_line("running")


def flush_caches(ns, flushcmd, flusharg, confirm):
    with ns.watch_log_from_here() as watcher:
        ns.rndc(f"{flushcmd} {flusharg}")
        watcher.wait_for_line(confirm)


def test_cacheclean_deleg(ns2):
    # Make sure the delegation cache has foo.bar.com.
    warmup_cache(ns2)

    # Flushing the cache
    check_cache(ns2, True)

    # Reloading the server keeps the cache hot
    reload_server(ns2)

    # The cache is still hot
    check_cache(ns2, True)

    # Flush the cache, and its now cold
    flush_caches(ns2, "flush", "", "flushing caches in all views succeeded")
    check_cache(ns2, False)

    # Flush just the name foo.bar.com.
    warmup_cache(ns2)
    check_cache(ns2, True)
    flush_caches(
        ns2,
        "flushname",
        "foo.bar.com.",
        "flushing name 'foo.bar.com.' in delegation cache for all views succeeded",
    )
    check_cache(ns2, False)

    # Flush the whole .com. tree (need to flush
    warmup_cache(ns2)
    check_cache(ns2, True)
    flush_caches(
        ns2,
        "flushtree",
        "com.",
        "flushing tree 'com.' in DNS cache for all views succeeded",
    )
    check_cache(ns2, False)

    # Check -expired
    warmup_cache(ns2)
    check_cache(ns2, True)
    check_cache_expired(ns2)



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from dns import flags

import pytest

from isctest.util import param

import isctest

pytestmark = pytest.mark.extra_artifacts(
    [
        "*/K*",
        "*/dsset-*",
        "*/*.bk",
        "*/*.conf",
        "*/*.db",
        "*/*.id",
        "*/*.jnl",
        "*/*.jbk",
        "*/*.key",
        "*/*.signed",
        "*/settime.out.*",
        "ans*/ans.run",
        "*/trusted.keys",
        "*/*.bad",
        "*/*.next",
        "*/*.stripped",
        "*/*.tmp",
        "*/*.stage?",
        "*/*.patched",
        "*/*.lower",
        "*/*.upper",
        "*/*.unsplit",
    ]
)


@pytest.mark.parametrize(
    "check, qname, qtype",
    [
        param("validation", "example.", "SOA"),
        param("negative-validation", "example.", "PTR"),
        param("insecurity-proof", "a.insecure.example.", "A"),
    ],
)
def test_misconfigured_ta_servfail(check, qname, qtype):
    isctest.log.info(f"check that {check} fails")
    msg = isctest.query.create(qname, qtype)
    res = isctest.query.tcp(msg, "10.53.0.5")
    isctest.check.servfail(res)


@pytest.mark.parametrize(
    "check, qname, qtype, rcode_func",
    [
        param("positive-answer", "example.", "SOA", isctest.check.noerror),
        param("negative-answer", "q.example.", "SOA", isctest.check.nxdomain),
        param("bogus-answer", "a.bogus.example.", "SOA", isctest.check.noerror),
        param("insecurity-proof", "a.insecure.example.", "SOA", isctest.check.noerror),
        param(
            "negative-insecurity-proof",
            "q.insecure.example.",
            "SOA",
            isctest.check.nxdomain,
        ),
    ],
)
def test_misconfigured_ta_with_cd(check, qname, qtype, rcode_func):
    isctest.log.info(f"check {check} with CD=1")
    msg = isctest.query.create(qname, qtype)
    msg.flags |= flags.CD
    res = isctest.query.tcp(msg, "10.53.0.5")
    rcode_func(res)
    isctest.check.noadflag(res)

    isctest.log.debug("compare the response from a correctly configured server")
    res2 = isctest.query.tcp(msg, "10.53.0.4")
    isctest.check.noadflag(res2)
    isctest.check.same_answer(res, res2)



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import pytest

pytestmark = pytest.mark.extra_artifacts(
    [
        "dig.out.*",
        "ns1/K*",
        "ns1/*.signed",
        "ns1/dsset-nsec3.",
        "ns1/dsset-signed.",
        "ns1/nsec3.db",
        "ns1/signed.db",
        "ns2/example.db",
        "ns2/named.stats",
        "ns2/redirect.db",
        "ns3/K*",
        "ns3/*.signed",
        "ns3/dsset-nsec3.",
        "ns3/dsset-signed.",
        "ns3/nsec3.db",
        "ns3/signed.db",
        "ns4/example.db",
        "ns4/named.stats",
        "ns5/K*",
        "ns5/dsset-*",
        "ns5/*.signed",
        "ns5/root.db",
        "ns5/sign.ns5.*",
        "ns5/signed.db",
        "ns6/signed.db.signed",
        "ns9/K*",
        "ns9/dsset-*",
        "ns9/root.db",
        "ns9/root.db.signed",
        "ns9/sign.ns9.*",
        "ns10/trusted.conf",
    ]
)


def test_redirect(run_tests_sh):
    run_tests_sh()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from re import compile as Re

import isctest


def query_and_dump(ns):
    msg = isctest.query.create("a.foo.test.", "A")
    res = isctest.query.udp(msg, ns.ip)
    isctest.check.noerror(res)

    msg = isctest.query.create("a.bar.test.", "A")
    res = isctest.query.udp(msg, ns.ip)
    isctest.check.noerror(res)

    with ns.watch_log_from_here() as watcher:
        ns.rndc("dumpdb -deleg")
        watcher.wait_for_line("dumpdb complete")
    return isctest.text.TextFile(f"{ns.identifier}/named_dump.db")


def test_cyclic_glues(ns1, ns4, templates):
    dump = query_and_dump(ns4)

    # The test is using the correctly-behaving ns2 server.
    assert len(dump.grep(Re("test. .* DELEG server-ipv4=10.53.0.2"))) == 1
    assert len(dump.grep(Re("test. .* DELEG server-ipv4=10.53.0.5"))) == 0

    # We've sent queries for both foo.test and bar.test and got a
    # single in-domain address from each:
    assert len(dump.grep(Re("foo.test. [0-9]* DELEG server-ipv4=10.53.0.3"))) == 1
    assert len(dump.grep(Re("foo.test. [0-9]* DELEG"))) == 1

    assert len(dump.grep(Re("bar.test. [0-9]* DELEG server-ipv4=10.53.0.3"))) == 1
    assert len(dump.grep(Re("bar.test. [0-9]* DELEG"))) == 1

    # in total we should have test, foo.test, and bar.test, nothing else:
    assert len(dump.grep(Re("test. [0-9]* DELEG"))) == 3

    templates.render("ns1/root.db", {"broken_ns": True})
    with ns1.watch_log_from_here() as watcher:
        ns1.rndc("reload")
        watcher.wait_for_all(["running", "zone ./IN: loaded serial"])

    with ns4.watch_log_from_here() as watcher:
        ns4.rndc("flush")
        watcher.wait_for_line("flushing caches in all views succeeded")

    dump = query_and_dump(ns4)

    # The test is now using the broken ans5 server.
    assert len(dump.grep(Re("test. [0-9]* DELEG server-ipv4=10.53.0.2"))) == 0
    assert len(dump.grep(Re("test. [0-9]* DELEG server-ipv4=10.53.0.5"))) == 1

    # The bar and foo delegations include names and glue records
    # that are out of bailiwick; we need to ensure that we are not
    # using the address, but only its name.
    assert len(dump.grep(Re("10.10.10.10"))) == 0
    assert len(dump.grep(Re("test2. "))) == 0
    assert len(dump.grep(Re("foo.test. [0-9]* DELEG server-name=ns.test2."))) == 1

    # There should in principle be only one of these, but there is no guard
    # to prevent duplicates when two glues for two different owner names
    # (in this case, ns.foo.test and ns.bar.test) both point to the same
    # IP address.
    assert len(dump.grep(Re("foo.test. [0-9]* DELEG server-ipv4="))) > 0

    # ns.bar.test is in-domain and should be stored as an address,
    # not a server-name.
    assert len(dump.grep(Re("bar.test. [0-9]* DELEG server-name=ns.bar.test."))) == 0
    assert len(dump.grep(Re("bar.test. [0-9]* DELEG server-ipv4=10.53.0.3"))) == 1

    # Since ns2.foo.test. came from the same parent (sibling glue) we have
    # its address, NOT its server-name.
    assert len(dump.grep(Re("bar.test. [0-9]* DELEG server-name=ns2.foo.test."))) == 0
    assert len(dump.grep(Re("bar.test. [0-9]* DELEG server-ipv4=10.53.0.4"))) == 1



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

"""
Fixtures shared by the digdelv test modules.
"""

from dataclasses import dataclass

import os

import pytest


@dataclass
class Zsk:
    keyid: str
    keydata: str
    rrcomment: str


@pytest.fixture(name="zsk")
def zsk_fixture():
    """Key id and rdata of the ZSK generated for the example zone."""
    with open("ns2/keyid", encoding="utf-8") as keyid_file:
        keyid = keyid_file.read().strip()
    with open("ns2/keydata", encoding="utf-8") as keydata_file:
        keydata = keydata_file.read().strip()
    return Zsk(
        keyid=keyid,
        keydata=keydata,
        rrcomment=f"; ZSK; alg = {os.environ['DEFAULT_ALGORITHM']} ; key id = {keyid}",
    )



"""
Copyright (C) Internet Systems Consortium, Inc. ("ISC")

SPDX-License-Identifier: MPL-2.0

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0.  If a copy of the MPL was not distributed with this
file, you can obtain one at https://mozilla.org/MPL/2.0/.

See the COPYRIGHT file distributed with this work for additional
information regarding copyright ownership.
"""

from collections.abc import AsyncGenerator

import dns.edns
import dns.name
import dns.rcode
import dns.rdatatype
import dns.rrset

from isctest.asyncserver import (
    AsyncDnsServer,
    DnsResponseSend,
    DomainHandler,
    IgnoreAllQueries,
    QnameHandler,
    QnameQtypeHandler,
    QueryContext,
    ResponseHandler,
    StaticResponseHandler,
)

from ..resolver_ans import (
    DelegationHandler,
    Gl6412AHandler,
    Gl6412Handler,
    Gl6412Ns2Handler,
    Gl6412Ns3Handler,
    rrset,
    setup_delegation,
    soa_rrset,
)


class BadGoodDnameNsHandler(QnameQtypeHandler, StaticResponseHandler):
    qnames = [
        "baddname.example.org.",
        "gooddname.example.org.",
    ]
    qtypes = [dns.rdatatype.NS]
    answer = [rrset("example.org.", dns.rdatatype.NS, "a.root-servers.nil.")]
    authoritative = True


def _cname_rrsets(
    qname: dns.name.Name | str,
) -> tuple[dns.rrset.RRset, dns.rrset.RRset]:
    return (
        rrset(qname, dns.rdatatype.CNAME, f"{qname}"),
        rrset(qname, dns.rdatatype.A, "1.2.3.4"),
    )


class Cname1Handler(QnameHandler, StaticResponseHandler):
    qnames = ["cname1.example.com."]
    # Data for the "cname + other data / 1" test
    answer = _cname_rrsets(qnames[0])
    authoritative = False


class Cname2Handler(QnameHandler, StaticResponseHandler):
    qnames = ["cname2.example.com."]
    # Data for the "cname + other data / 2" test: same RRs in opposite order
    answer = tuple(reversed(_cname_rrsets(qnames[0])))
    authoritative = False


class ExampleOrgHandler(QnameHandler):
    qnames = [
        "www.example.org",
        "badcname.example.org",
        "goodcname.example.org",
        "foo.baddname.example.org",
        "foo.gooddname.example.org",
    ]

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        # Data for address/alias filtering.
        if qctx.qtype == dns.rdatatype.A:
            a_rrset = rrset(qctx.qname, dns.rdatatype.A, "192.0.2.1")
            qctx.response.answer.append(a_rrset)
        elif qctx.qtype == dns.rdatatype.AAAA:
            aaaa_rrset = rrset(qctx.qname, dns.rdatatype.AAAA, "2001:db8:beef::1")
            qctx.response.answer.append(aaaa_rrset)
        yield DnsResponseSend(qctx.response, authoritative=True)


class NoResponseExampleUdpHandler(QnameHandler, IgnoreAllQueries):
    qnames = ["noresponse.exampleudp.net."]


class RootNsHandler(QnameQtypeHandler):
    qnames = [
        "example.com.",
        "com.",
        "example.org.",
        "org.",
        "net.",
    ]

    qtypes = [dns.rdatatype.NS]

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        root_ns = rrset(qctx.qname, dns.rdatatype.NS, "a.root-servers.nil.")
        qctx.response.answer.append(root_ns)
        yield DnsResponseSend(qctx.response, authoritative=True)


class ZoneVersionHandler(QnameHandler):
    qnames = ["zoneversion."]

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        qctx.response.authority.append(soa_rrset("."))
        zoneversion_opt = dns.edns.GenericOption(19, bytes.fromhex("000101022304"))  # type: ignore
        qctx.response.use_edns(edns=0, options=[zoneversion_opt])
        yield DnsResponseSend(qctx.response, authoritative=False)


class Ns2Delegation(DelegationHandler):
    domains = ["exampleudp.net."]
    server_number = 2


class Ns3Delegation(DelegationHandler):
    domains = [
        "example.net.",
        "isc.org.",
        "lame.example.org.",
        "sub.example.org.",
    ]
    server_number = 3


class Ns3GlueInAnswerDelegation(DelegationHandler):
    domains = ["glue-in-answer.example.org."]
    server_number = 3

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        async for dns_response in super().get_responses(qctx):
            dns_response.response.answer += dns_response.response.additional
            yield dns_response


class Ns4Delegation(DelegationHandler):
    domains = ["broken."]
    server_number = 4


class Ns6Delegation(DelegationHandler):
    domains = [
        "redirect.com.",
        "tld1.",
    ]
    server_number = 6


class Ns7Delegation(DelegationHandler):
    domains = ["tld2."]
    server_number = 7


class PartialFormerrHandler(DomainHandler, StaticResponseHandler):
    domains = ["partial-formerr."]
    authoritative = False
    rcode = dns.rcode.FORMERR


class FallbackHandler(ResponseHandler):
    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        setup_delegation(qctx, "below.www.example.com.", 3)
        yield DnsResponseSend(qctx.response, authoritative=False)


def main() -> None:
    server = AsyncDnsServer(default_rcode=dns.rcode.NOERROR)

    # Install QnameHandlers first
    server.install_response_handlers(
        BadGoodDnameNsHandler(),
        Cname1Handler(),
        Cname2Handler(),
        ExampleOrgHandler(),
        Gl6412AHandler(),
        Gl6412Handler(),
        Gl6412Ns2Handler(),
        Gl6412Ns3Handler(),
        NoResponseExampleUdpHandler(),
        RootNsHandler(),
        ZoneVersionHandler(),
    )

    # Then install DomainHandlers
    server.install_response_handlers(
        Ns2Delegation(),
        Ns3Delegation(),
        Ns3GlueInAnswerDelegation(),
        Ns4Delegation(),
        Ns6Delegation(),
        Ns7Delegation(),
        PartialFormerrHandler(),
    )

    # Finally, install the fallback handler
    server.install_response_handler(FallbackHandler())
    server.run()


if __name__ == "__main__":
    main()



"""
Copyright (C) Internet Systems Consortium, Inc. ("ISC")

SPDX-License-Identifier: MPL-2.0

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0.  If a copy of the MPL was not distributed with this
file, you can obtain one at https://mozilla.org/MPL/2.0/.

See the COPYRIGHT file distributed with this work for additional
information regarding copyright ownership.
"""

from collections.abc import AsyncGenerator

import struct

import dns.flags
import dns.rcode
import dns.rdatatype

from isctest.asyncserver import (
    AsyncDnsServer,
    BytesResponseSend,
    DnsProtocol,
    DnsResponseSend,
    QueryContext,
    ResponseAction,
    ResponseHandler,
)

# DNS constants used by raw wire builder functions below
DNS_TYPE_SOA = 6
DNS_TYPE_A = 1
DNS_TYPE_NS = 2
DNS_TYPE_AXFR = 252
DNS_TYPE_IXFR = 251
DNS_CLASS_IN = 1
DNS_FLAG_QR = 0x8000
DNS_FLAG_AA = 0x0400
DNS_RCODE_NOERROR = 0
DNS_RCODE_SERVFAIL = 2

ZONE_NAME = "ixfr-race."
NUM_RECORDS = 400


def encode_name(name):
    """Encode a DNS name in wire format (no compression)."""
    parts = name.rstrip(".").split(".")
    result = b""
    for part in parts:
        encoded = part.encode("ascii")
        result += struct.pack("B", len(encoded)) + encoded
    result += b"\x00"
    return result


def build_soa_rdata(
    mname, rname, serial, refresh=3600, retry=900, expire=604800, minimum=86400
):
    """Build SOA record rdata."""
    rdata = encode_name(mname)
    rdata += encode_name(rname)
    rdata += struct.pack("!IIIII", serial, refresh, retry, expire, minimum)
    return rdata


def build_a_rdata(ip_str):
    """Build A record rdata from dotted-quad string."""
    parts = ip_str.split(".")
    return struct.pack("4B", *[int(p) for p in parts])


def build_rr(name_bytes, rtype, rclass, ttl, rdata):
    """Build a complete resource record."""
    rr = name_bytes
    rr += struct.pack("!HHIH", rtype, rclass, ttl, len(rdata))
    rr += rdata
    return rr


def build_dns_header(qid, flags, qdcount, ancount, nscount=0, arcount=0):
    """Build DNS message header."""
    return struct.pack("!HHHHHH", qid, flags, qdcount, ancount, nscount, arcount)


def build_ixfr_message1(qid, zone_name, num_records):
    """
    Build IXFR Message 1: A valid IXFR diff that triggers ixfr_commit().

    This message contains a complete diff 1 (large, many records) which
    triggers ixfr_commit() -> isc_work_enqueue() -> worker thread starts.

    The message ends with a boundary SOA that starts diff 2, so the state
    machine is in XFRST_IXFR_DEL waiting for more records.

    Answer section structure:
    1. Initial SOA (end_serial=3)   -- XFRST_ZONEXFRREQUEST
    2. Old SOA (serial=1)           -- XFRST_FIRSTDATA -> IXFR -> DELSOA
    3. DEL A records (num_records)  -- XFRST_IXFR_DEL (diffs++)
    4. Mid SOA (serial=2)           -- XFRST_IXFR_ADDSOA (diffs++)
    5. ADD A records (num_records)  -- XFRST_IXFR_ADD (diffs++)
    6. Boundary SOA (serial=2)      -- ixfr_commit()! Worker enqueued.
                                       Then goto redo -> DELSOA of diff 2
    """
    zone_wire = encode_name(zone_name)
    question = zone_wire + struct.pack("!HH", DNS_TYPE_IXFR, DNS_CLASS_IN)

    mname = "ns." + zone_name
    rname = "admin." + zone_name
    end_serial = 3
    old_serial = 1
    mid_serial = 2

    soa_end = build_soa_rdata(mname, rname, end_serial)
    soa_old = build_soa_rdata(mname, rname, old_serial)
    soa_mid = build_soa_rdata(mname, rname, mid_serial)

    records = []

    # 1. Initial SOA (end serial)
    records.append(build_rr(zone_wire, DNS_TYPE_SOA, DNS_CLASS_IN, 3600, soa_end))

    # 2. Old SOA (serial 1) - triggers IXFR detection
    records.append(build_rr(zone_wire, DNS_TYPE_SOA, DNS_CLASS_IN, 3600, soa_old))

    # 3. DEL A records
    for i in range(num_records):
        name = encode_name(f"host-{i}.{zone_name}")
        ip = f"10.0.{(i >> 8) & 0xFF}.{i & 0xFF}"
        records.append(
            build_rr(name, DNS_TYPE_A, DNS_CLASS_IN, 3600, build_a_rdata(ip))
        )

    # 4. Mid SOA (serial 2) - end of DEL, start of ADD
    records.append(build_rr(zone_wire, DNS_TYPE_SOA, DNS_CLASS_IN, 3600, soa_mid))

    # 5. ADD A records
    for i in range(num_records):
        name = encode_name(f"host-{i}.{zone_name}")
        ip = f"10.1.{(i >> 8) & 0xFF}.{i & 0xFF}"
        records.append(
            build_rr(name, DNS_TYPE_A, DNS_CLASS_IN, 3600, build_a_rdata(ip))
        )

    # 6. Boundary SOA (serial=2 == current_serial) -> ixfr_commit()!
    # This triggers the worker thread via isc_work_enqueue().
    # Then goto redo processes it as DELSOA of diff 2.
    records.append(build_rr(zone_wire, DNS_TYPE_SOA, DNS_CLASS_IN, 3600, soa_mid))

    ancount = len(records)
    answer = b"".join(records)
    flags = DNS_FLAG_QR | DNS_FLAG_AA | DNS_RCODE_NOERROR
    header = build_dns_header(qid, flags, 1, ancount)
    msg = header + question + answer

    return msg


def build_bad_rcode_message2(qid, zone_name):
    """
    Build Message 2

    A DNS response with rcode=SERVFAIL. When BIND receives this during an
    active IXFR transfer:

    xfrin_recv_done():
      msg->rcode != dns_rcode_noerror  (SERVFAIL != NOERROR) ->
      result = dns_result_fromrcode(msg->rcode) ->
      reqtype == dns_rdatatype_ixfr (not axfr/soa) ->
      falls through to try_axfr: ->
      xfrin_reset() -> destroys journal/version

    Meanwhile ixfr_apply worker from Message 1 is still running -> UAF.

    This works with DEFAULT secondary configuration (no special options).
    """
    zone_wire = encode_name(zone_name)
    question = zone_wire + struct.pack("!HH", DNS_TYPE_IXFR, DNS_CLASS_IN)

    flags = DNS_FLAG_QR | DNS_FLAG_AA | DNS_RCODE_SERVFAIL
    header = build_dns_header(qid, flags, 1, 0)
    msg = header + question

    return msg


def build_soa_response(qid, zone_name, serial):
    """Build a SOA response for the zone."""
    zone_wire = encode_name(zone_name)
    question = zone_wire + struct.pack("!HH", DNS_TYPE_SOA, DNS_CLASS_IN)

    mname = "ns." + zone_name
    rname = "admin." + zone_name
    soa_rdata = build_soa_rdata(mname, rname, serial)
    answer = build_rr(zone_wire, DNS_TYPE_SOA, DNS_CLASS_IN, 3600, soa_rdata)

    flags = DNS_FLAG_QR | DNS_FLAG_AA | DNS_RCODE_NOERROR
    header = build_dns_header(qid, flags, 1, 1)
    return header + question + answer


def build_axfr_response(qid, zone_name, serial, num_records):
    """
    Build a complete AXFR response for initial zone load.

    AXFR format: SOA, NS, A records, ..., SOA (trailing SOA marks end).
    """
    zone_wire = encode_name(zone_name)
    question = zone_wire + struct.pack("!HH", DNS_TYPE_AXFR, DNS_CLASS_IN)

    mname = "ns." + zone_name
    rname = "admin." + zone_name
    soa_rdata = build_soa_rdata(mname, rname, serial)

    records = []

    # Opening SOA
    records.append(build_rr(zone_wire, DNS_TYPE_SOA, DNS_CLASS_IN, 3600, soa_rdata))

    # NS record
    ns_wire = encode_name("ns." + zone_name)
    records.append(build_rr(zone_wire, DNS_TYPE_NS, DNS_CLASS_IN, 3600, ns_wire))

    # NS A record
    records.append(
        build_rr(ns_wire, DNS_TYPE_A, DNS_CLASS_IN, 3600, build_a_rdata("127.0.0.1"))
    )

    # A records (matching gen_zone.py output)
    for i in range(num_records):
        name = encode_name(f"host-{i}.{zone_name}")
        ip = f"10.0.{(i >> 8) & 0xFF}.{i & 0xFF}"
        records.append(
            build_rr(name, DNS_TYPE_A, DNS_CLASS_IN, 3600, build_a_rdata(ip))
        )

    # Trailing SOA (marks end of AXFR)
    records.append(build_rr(zone_wire, DNS_TYPE_SOA, DNS_CLASS_IN, 3600, soa_rdata))

    ancount = len(records)
    answer = b"".join(records)
    flags = DNS_FLAG_QR | DNS_FLAG_AA | DNS_RCODE_NOERROR
    header = build_dns_header(qid, flags, 1, ancount)
    msg = header + question + answer

    return msg


class IxfrRaceHandler(ResponseHandler):
    """
    Handle SOA, AXFR, and IXFR queries to trigger the IXFR->AXFR race condition.

    Phase 1: Respond to SOA with serial=1 and serve an AXFR to load the zone.
    Phase 2: After AXFR, respond to SOA with serial=3 to trigger IXFR.
             On IXFR, send a valid large diff (msg1) followed immediately by a
             SERVFAIL response (msg2) to race ixfr_commit() against xfrin_reset().
    """

    def __init__(self) -> None:
        self._axfr_done = False

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        qid = qctx.query.id

        if qctx.qtype == dns.rdatatype.SOA:
            serial = 3 if self._axfr_done else 1
            yield BytesResponseSend(build_soa_response(qid, ZONE_NAME, serial))

        elif qctx.qtype == dns.rdatatype.AXFR:
            yield BytesResponseSend(build_axfr_response(qid, ZONE_NAME, 1, NUM_RECORDS))
            self._axfr_done = True

        elif qctx.qtype == dns.rdatatype.IXFR:
            if qctx.protocol == DnsProtocol.UDP:
                # Force TCP retry by setting the TC bit
                qctx.response.flags |= dns.flags.TC
                yield DnsResponseSend(qctx.response)
            else:
                # Message 1: Valid IXFR diff -> triggers ixfr_commit()
                yield BytesResponseSend(
                    build_ixfr_message1(qid, ZONE_NAME, NUM_RECORDS)
                )
                # Message 2: SERVFAIL -> triggers xfrin_reset() while
                # ixfr_apply worker from Message 1 is still running -> UAF
                yield BytesResponseSend(build_bad_rcode_message2(qid, ZONE_NAME))


def main() -> None:
    server = AsyncDnsServer(default_rcode=dns.rcode.NOERROR, default_aa=True)
    server.install_response_handler(IxfrRaceHandler())
    server.run()


if __name__ == "__main__":
    main()



"""
Copyright (C) Internet Systems Consortium, Inc. ("ISC")

SPDX-License-Identifier: MPL-2.0

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0.  If a copy of the MPL was not distributed with this
file, you can obtain one at https://mozilla.org/MPL/2.0/.

See the COPYRIGHT file distributed with this work for additional
information regarding copyright ownership.
"""

from collections.abc import AsyncGenerator

import dns.rcode

from isctest.asyncserver import (
    ControllableAsyncDnsServer,
    DnsResponseSend,
    QueryContext,
)

from ..reclimit_ans import (
    DirectExampleHandler,
    FallbackNxdomainHandler,
    IndirectExampleOrgHandler,
    LimitControlCommand,
    Ns1ExampleOrgHandler,
    ReclimitHandler,
    ReclimitStateHandler,
    a,
    is_ns1_example,
    ns,
)


class Ns1ExampleNetHandler(ReclimitHandler):
    def match(self, qctx: QueryContext) -> bool:
        return is_ns1_example(qctx.qname, "net")

    async def _get_counted_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        current_ns_number = int(qctx.qname.labels[1])
        next_ns_block_start = (current_ns_number + 1) * 16
        for offset in range(1, 16):
            target_ns_number = next_ns_block_start + offset
            qctx.response.authority.append(
                ns(
                    f"{current_ns_number}.example.net.",
                    f"ns1.{target_ns_number}.example.net.",
                )
            )
            qctx.response.additional.append(
                a(f"ns1.{target_ns_number}.example.net.", 7)
            )

        yield DnsResponseSend(qctx.response, authoritative=False)


def main() -> None:
    server = ControllableAsyncDnsServer(
        default_aa=True, default_rcode=dns.rcode.NOERROR
    )
    server.install_response_handlers(
        state_handler := ReclimitStateHandler(indirect_send_response_default=False),
        DirectExampleHandler(state_handler, 2),
        IndirectExampleOrgHandler(state_handler, 2),
        Ns1ExampleOrgHandler(state_handler),
        Ns1ExampleNetHandler(state_handler),
        FallbackNxdomainHandler(state_handler),
    )
    server.install_control_command(LimitControlCommand(state_handler))
    server.run()


if __name__ == "__main__":
    main()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

"""
Tests for the look-up tools other than dig, delv and mdig: nslookup,
host and nsupdate.
"""

from textwrap import dedent

import pytest

from digdelv.common import ARTIFACTS

import isctest

pytestmark = pytest.mark.extra_artifacts(ARTIFACTS)


def test_nslookup_update_response(named_port, ans6):
    """Check that nslookup rejects a response with the UPDATE opcode."""
    nslookup = isctest.run.EnvCmd("NSLOOKUP")
    result = nslookup(
        f"-port={named_port} -q=CNAME -timeout=1 foo.bar {ans6.ip}",
        raise_on_exception=False,
    )
    assert result.rc != 0
    assert "Opcode mismatch" in result.out


def test_host_update_response(named_port, ans6):
    """Check that host rejects a response with the UPDATE opcode."""
    host = isctest.run.EnvCmd("HOST")
    result = host(
        f"-p {named_port} -W 1 -t CNAME foo.bar {ans6.ip}", raise_on_exception=False
    )
    assert result.rc != 0
    assert "Opcode mismatch" in result.out


def test_nsupdate_update_response(named_port, ans6):
    """Check that nsupdate rejects an UPDATE response to its SOA query."""
    nsupdate = isctest.run.EnvCmd("NSUPDATE")
    commands = dedent(f"""\
        server {ans6.ip} {named_port}
        add x.example.com 300 in a 1.2.3.4
        send
    """)
    result = nsupdate("", input_text=commands.encode(), raise_on_exception=False)
    assert result.rc == 1
    assert "invalid OPCODE in response to SOA query" in result.err



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

"""
Tests for the mdig tool.
"""

import os

import pytest

from digdelv.common import ARTIFACTS, parse_yaml

import isctest

pytestmark = [
    pytest.mark.extra_artifacts(ARTIFACTS),
]


@pytest.fixture(name="mdig")
def mdig_fixture(named_port):
    return isctest.run.EnvCmd("MDIG", f"-p {named_port}")


def test_tcp_with_source_address_and_port(mdig, ns3):
    """Check that mdig +tcp works with a source address and port.  When
    running more than once in quick succession with a source
    address#port, the query can fail with "response failed with address
    not available" because the address#port is still busy; that error is
    not interesting, only the unexpected error case is.  See GL #4969
    for more information."""
    extraport8 = os.environ["EXTRAPORT8"]
    result = mdig(
        f"-b {ns3.ip}#{extraport8} +tcp @{ns3.ip} example", raise_on_exception=False
    )
    assert "unexpected error" not in result.out
    assert "unexpected error" not in result.err


def test_ednsopt_malformed(mdig, ns3):
    """Check that mdig handles the malformed option '+ednsopt=:'
    gracefully."""
    result = mdig(f"@{ns3.ip} +ednsopt=: a.example", raise_on_exception=False)
    assert result.rc != 0
    assert "ednsopt no code point specified" in result.err


def test_dnskey_norrcomments(mdig, ns3, zsk):
    """Check that +multi +norrcomments suppresses the DNSKEY comment
    (the default is rrcomments)."""
    result = mdig(f"+tcp @{ns3.ip} +multi +norrcomments -t DNSKEY example")
    assert zsk.rrcomment not in result.out


def test_soa_norrcomments(mdig, ns3):
    """Check that +multi +norrcomments suppresses the SOA field
    comments."""
    result = mdig(f"+tcp @{ns3.ip} +multi +norrcomments -t SOA example")
    assert "; serial" not in result.out


def test_yaml_output(mdig, ns3):
    """Check the structure of mdig +yaml output."""
    result = mdig(f"+yaml @{ns3.ip} -t any ns2.example")
    response = parse_yaml(result.out)[0]["message"]["response_message_data"]
    assert response["status"] == "NOERROR"
    assert response["QUESTION_SECTION"][0] == "ns2.example. IN ANY"



"""
Copyright (C) Internet Systems Consortium, Inc. ("ISC")

SPDX-License-Identifier: MPL-2.0

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0.  If a copy of the MPL was not distributed with this
file, you can obtain one at https://mozilla.org/MPL/2.0/.

See the COPYRIGHT file distributed with this work for additional
information regarding copyright ownership.
"""

from collections.abc import AsyncGenerator
from typing import final

import abc
import asyncio

import dns.flags
import dns.name
import dns.rcode
import dns.rdataclass
import dns.rdatatype
import dns.rrset

from isctest.asyncserver import (
    ControlCommand,
    ControllableAsyncDnsServer,
    DnsResponseSend,
    QnameHandler,
    QueryContext,
    ResponseAction,
    ResponseHandler,
)


class ReclimitStateHandler(QnameHandler):
    """
    Handler for the "count." and "reset." queries that also holds the state
    shared by all the handlers in one server.
    """

    qnames = ["count.", "reset."]

    def __init__(self, indirect_send_response_default: bool = True) -> None:
        self._indirect_send_response_default = indirect_send_response_default
        self.count = 0
        self.limit = 0
        self.indirect_send_response = indirect_send_response_default
        super().__init__()

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        if f"{qctx.qname}" == "count.":
            self.count += 1
            qctx.response.answer.append(
                dns.rrset.from_text(
                    "count.", 0, dns.rdataclass.IN, dns.rdatatype.TXT, f"{self.count}"
                )
            )
            yield DnsResponseSend(qctx.response, authoritative=True)
        elif f"{qctx.qname}" == "reset.":
            self.reset()
            yield DnsResponseSend(qctx.response, authoritative=False)

    def reset(self) -> None:
        self.count = 0
        self.indirect_send_response = self._indirect_send_response_default


class ReclimitHandler(ResponseHandler):
    """
    Base class for handlers in this test.

    Increments the shared query counter on each query and delegates the actual
    response generation to the `_get_counted_responses()` method.
    """

    def __init__(self, state_handler: ReclimitStateHandler) -> None:
        self._state = state_handler
        super().__init__()

    @final
    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        self._state.count += 1
        async for response in self._get_counted_responses(qctx):
            yield response

    @abc.abstractmethod
    async def _get_counted_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        yield DnsResponseSend(qctx.response)


class LimitControlCommand(ControlCommand):
    control_subdomain = "limit"

    def __init__(self, state_handler: ReclimitStateHandler) -> None:
        self._state_handler = state_handler
        super().__init__()

    def handle(
        self, args: list[str], server: ControllableAsyncDnsServer, qctx: QueryContext
    ) -> str | None:
        if len(args) != 1:
            return "Expected exactly one label"

        try:
            limit = int(args[0])
        except ValueError:
            return "Expected an integer"

        self._state_handler.limit = limit
        return f"Limit set to {limit}"


def a(owner: str | dns.name.Name, ns_number: int) -> dns.rrset.RRset:
    return dns.rrset.from_text(
        f"{owner}", 3600, dns.rdataclass.IN, dns.rdatatype.A, f"10.53.0.{ns_number}"
    )


def ns(owner: str | dns.name.Name, target: str | dns.name.Name) -> dns.rrset.RRset:
    return dns.rrset.from_text(
        f"{owner}", 86400, dns.rdataclass.IN, dns.rdatatype.NS, f"{target}"
    )


class DirectExampleHandler(ReclimitHandler, QnameHandler):
    qnames = ["direct.example.org", "direct.example.net"]

    def __init__(
        self, state_handler: ReclimitStateHandler, local_ns_number: int
    ) -> None:
        self._local_ns_number = local_ns_number
        super().__init__(state_handler)

    async def _get_counted_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        if qctx.qtype == dns.rdatatype.A:
            qctx.response.answer.append(a(qctx.qname, self._local_ns_number))
        yield DnsResponseSend(qctx.response)


class IndirectExampleOrgHandler(ReclimitHandler, QnameHandler):
    qnames = [f"indirect{i}.example.org" for i in range(1, 9)]

    def __init__(
        self, state_handler: ReclimitStateHandler, local_ns_number: int
    ) -> None:
        self._local_ns_number = local_ns_number
        super().__init__(state_handler)

    async def _get_counted_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        if not self._state.indirect_send_response:
            qctx.response.authority.append(ns(f"{qctx.qname}", "ns1.1.example.org."))
            qctx.response.flags &= ~dns.flags.AA
        elif qctx.qtype == dns.rdatatype.A:
            qctx.response.answer.append(a(qctx.qname, self._local_ns_number))
        yield DnsResponseSend(qctx.response)


def is_ns1_example(qname: dns.name.Name, tld: str) -> bool:
    labels = qname.labels
    return (
        len(labels) == 5
        and labels[3] == tld.encode()
        and labels[2] == b"example"
        and labels[1].isdigit()
        and labels[0] == b"ns1"
    )


class Ns1ExampleOrgHandler(ReclimitHandler):
    def __init__(self, state_handler: ReclimitStateHandler) -> None:
        self._second_query_events: dict[dns.name.Name, asyncio.Event] = {}
        super().__init__(state_handler)

    def match(self, qctx: QueryContext) -> bool:
        return is_ns1_example(qctx.qname, "org") and qctx.qtype in (
            dns.rdatatype.A,
            dns.rdatatype.AAAA,
        )

    async def _get_counted_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        ns_number = int(qctx.qname.labels[1])
        next_ns_number = ns_number + 1
        if not self._state.limit or (
            not self._state.indirect_send_response
            and next_ns_number <= self._state.limit
        ):
            qctx.response.authority.append(
                ns(f"{ns_number}.example.org.", f"ns1.{next_ns_number}.example.org.")
            )
            qctx.response.flags &= ~dns.flags.AA
        else:
            self._state.indirect_send_response = True
            if qctx.qtype == dns.rdatatype.A:
                qctx.response.answer.append(a(qctx.qname, 4))

        second_query_event = self._second_query_events.get(qctx.qname)
        if second_query_event is not None:
            # Second query arrived, release the first response.
            second_query_event.set()
            await asyncio.sleep(0)  # Yield to allow the first response to be sent.
            yield DnsResponseSend(qctx.response)
        else:
            # Delay the response until the second query for the same QNAME
            # arrives; give up waiting after 500 ms.
            second_query_event = asyncio.Event()
            self._second_query_events[qctx.qname] = second_query_event
            try:
                await asyncio.wait_for(second_query_event.wait(), timeout=0.5)
            except asyncio.TimeoutError:
                pass
            finally:
                del self._second_query_events[qctx.qname]
            yield DnsResponseSend(qctx.response)


class FallbackNxdomainHandler(ReclimitHandler):
    async def _get_counted_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[DnsResponseSend, None]:
        qctx.response.set_rcode(dns.rcode.NXDOMAIN)
        yield DnsResponseSend(qctx.response)



"""
Copyright (C) Internet Systems Consortium, Inc. ("ISC")

SPDX-License-Identifier: MPL-2.0

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0.  If a copy of the MPL was not distributed with this
file, you can obtain one at https://mozilla.org/MPL/2.0/.

See the COPYRIGHT file distributed with this work for additional
information regarding copyright ownership.
"""

import dns.rcode
import dns.rdataclass
import dns.rdatatype
import dns.rrset

from isctest.asyncserver import AsyncDnsServer, DomainHandler, StaticResponseHandler


class PoisonReferralForwarder(DomainHandler, StaticResponseHandler):
    domains = ["fwd.hack."]
    authority = [
        dns.rrset.from_text(
            "hack.", 300, dns.rdataclass.IN, dns.rdatatype.NS, "ns.fwd.hack."
        )
    ]
    additional = [
        dns.rrset.from_text(
            "ns.fwd.hack.", 300, dns.rdataclass.IN, dns.rdatatype.A, "10.53.0.3"
        )
    ]
    rcode = dns.rcode.NOERROR
    authoritative = False


def main() -> None:
    server = AsyncDnsServer()
    server.install_response_handler(PoisonReferralForwarder())
    server.run()


if __name__ == "__main__":
    main()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from re import compile as Re

import dns.message

import isctest


# This test verifies the query pattern when the upstream behaves badly.
# In this scenario, the upstream server (ans3) always responds with a
# BADCOOKIE error for queries within the "example" zone, even on TCP.
# The resolver (ns4), should not resend the same queries over and over
# again, up to the max-query-count threshold. Instead, the expected
# pattern is:
# 1. Priming query, getting the NS for .
# 2. Getting the NS for example.
# 3. Trying to resolve test.example.
# 4. Trying again, but now with the server cookie.
# 5. Trying again, now over TCP.
#
# This means we expect 5 recursion queries trying to resolve test.example.
def test_resend_loop_badcookie(ns4):
    sending_packet = Re("sending packet from 10.53.0.4#[0-9]+ to 10.53.0.3#[0-9]+")
    received_packet = Re("received packet from 10.53.0.3#[0-9]+ to 10.53.0.4#[0-9]+")

    log_sequence = [
        # 1. Priming query, getting the NS for .
        sending_packet,
        Re("COOKIE: [0-9a-z]{16}$"),
        Re(".\\s+IN\\s+NS"),
        # 2. Getting the NS for example.
        sending_packet,
        Re("COOKIE: [0-9a-z]{16}$"),
        Re("example.\\s+IN\\s+NS"),
        # 3. Trying to resolve test.example.
        sending_packet,
        Re("COOKIE: [0-9a-z]{16}$"),
        Re("test.example.\\s+IN\\s+A"),
        # Get the first BADCOOKIE error.
        "UDP response",
        received_packet,
        "BADCOOKIE",
        Re("COOKIE: [0-9a-z]{16}1122334455667788"),
        Re("test.example.\\s+IN\\s+A"),
        # 4. Trying again, but now with the server cookie.
        sending_packet,
        Re("test.example.\\s+IN\\s+A"),
        # Get BADCOOKIE error again.
        "UDP response",
        received_packet,
        "BADCOOKIE",
        Re("COOKIE: [0-9a-z]{16}1122334455667788"),
        Re("test.example.\\s+IN\\s+A"),
        # 5. Trying again, now over TCP.
        sending_packet,
        Re("test.example.\\s+IN\\s+A"),
        # Fails and give up.
        "TCP response",
        received_packet,
        "BADCOOKIE",
        Re("COOKIE: [0-9a-z]{16}1122334455667788"),
        Re("test.example.\\s+IN\\s+A"),
    ]

    msg = dns.message.make_query("test.example", "A")
    with ns4.watch_log_from_here() as watcher:
        res = isctest.query.udp(msg, ns4.ip)
        watcher.wait_for_sequence(log_sequence)

    assert len(ns4.log.grep(sending_packet)) == 5

    isctest.check.servfail(res)

    prohibited_log = "query failed (timed out) for test.example/IN/A"
    assert prohibited_log not in ns4.log


def test_resend_loop_forward(ns1):
    sending_packet = Re("sending packet from 10.53.0.1#[0-9]+ to 10.53.0.3#[0-9]+")
    received_packet = Re("received packet from 10.53.0.3#[0-9]+ to 10.53.0.1#[0-9]+")

    # Make sure the server is done with priming and ./DNSKEY refresh (RFC5011)
    msg = dns.message.make_query(".", "NS")
    isctest.query.udp(msg, ns1.ip)
    query_count_start = len(ns1.log.grep(sending_packet))

    log_sequence = [
        sending_packet,
        "flags: rd; QUESTION: 1",
        received_packet,
        "opcode: QUERY, status: SERVFAIL",
        "flags: qr rd; QUESTION: 1",
        sending_packet,
        "flags: rd cd; QUESTION: 1",
        received_packet,
        "opcode: QUERY, status: SERVFAIL",
        "flags: qr rd; QUESTION: 1",
    ]

    msg = dns.message.make_query("test.com.", "A")
    with ns1.watch_log_from_here() as watcher:
        res = isctest.query.udp(msg, ns1.ip)
        watcher.wait_for_sequence(log_sequence)

    isctest.check.servfail(res)

    query_count_end = len(ns1.log.grep(sending_packet))
    assert query_count_end - query_count_start == 2

    prohibited_log = "query failed (timed out) for test.com/IN/A"
    assert prohibited_log not in ns1.log
